"""Top-level service organizing databases."""

import glob
import os.path

from . import DEFAULT_POCKET_NAME, exceptions, init_logger, pocket

logger = init_logger(__name__)


class Server:
    """Server class holding the ``TinyDbPocket`` databases.

    All database handling is taken care of in the underlying `TinyDbPocket`.
    Kwargs (f.i. storage) are passed to the TinyDbPocket member.
    """

    def __init__(self, **kwargs):
        self._pockets = {}
        self._pocket_kwargs = kwargs

    def run(self, command, **kwargs):
        """The requested pocket is created if not yet present. The method of
        `Pocket` corresponding to the given `command` is called. All `kwargs`
        are passed on. A json-like response is returned.

        Wrap this in a 'broad' try-except block to catch any server-side errors.
        :return: dict
            key is one of 'id', 'element', 'elements', 'error', 'pockets'
        """
        logger.debug(f"Running '{command}' with {kwargs}")

        try:
            if command == "pockets":
                return {"pockets": self._pocket_names()}
            elif command == "copy":
                return {"id": self._copy_entry(**kwargs)}
            elif command == "stop":
                # graceful shutdown, invoke closing of files
                for pd in self._pockets.values():
                    pd.close()
                return {}
            else:
                pocket_name = kwargs.pop("pocket", None)
                pd = self._get_pocket(pocket_name)

                if command == "add":
                    response = {"id": pd.add_entry(**kwargs)}
                elif command == "remove":
                    response = {"id": pd.remove_entry(**kwargs)}
                elif command == "list":
                    response = {"elements": pd.get_entries(**kwargs)}
                elif command == "get":
                    response = {"element": pd.get_entry(**kwargs)}
                elif command == "update":
                    response = {"id": pd.update_entry(**kwargs)}
                elif command == "categories":
                    response = {"categories": pd.get_categories()}
                else:
                    response = {"error": f"Server: unknown command '{command}'"}
                return response

        except exceptions.PocketException as e:
            return {"error": e}

    def _get_pocket(self, name=None):
        """Get the Pocket identified by 'name' from the Pockets dictionary. If
        the Pocket does not exist, it is created and returned. If 'name' is
        None, the default pocket name is used as defined in __init__.py

        :type name: str or None
        :return: Pocket object
        """
        name = name or DEFAULT_POCKET_NAME
        try:
            pd = self._pockets[name]
        except KeyError:
            logger.debug(f"Loading pocket '{name}'")
            pd = pocket.TinyDbPocket(name, **self._pocket_kwargs)
            self._pockets[pd.name] = pd

        return pd

    def _pocket_names(self):
        """Return names of pockets currently organized by the server.
        If persistent data storage was specified, all JSON files present in the
        'data_dir' are also taken into account.

        :return: list(str)
        """
        # Avoid duplicate entries if Pocket already present in self._pockets by
        # using sets
        names = {p._name for p in self._pockets.values()}

        data_dir = self._pocket_kwargs.get("data_dir")
        names.update(pocket_names(data_dir))
        return sorted(names)

    def _copy_entry(self, source_pocket=None, destination_pocket=None, **kwargs):
        """Copy an entry (specified by ID and table_name) from the source pocket
        to the destination pocket.

        :type _pocket: str
        :return: ID of copied entry
        :raises: PocketException if the source entry does not exist
        """
        source_pocket = self._get_pocket(source_pocket)
        entry_to_copy = source_pocket.get_entry(**kwargs)

        destination_pocket = self._get_pocket(destination_pocket)
        return destination_pocket.add_entry(
            table_name=kwargs.get("table_name"), **entry_to_copy
        )


def pocket_names(data_dir):
    """Return names of all pockets (i.e. names of JSON files in the given data
    directory), or an empty set if the specified data directory is None.
    """
    if data_dir is None:
        return set()
    result = {
        os.path.splitext(os.path.basename(f))[0]
        for f in glob.glob(os.path.join(data_dir, "*.json"))
    }
    return result



"""Local server proxy for direct communication (client and server reside in
common process)."""

from . import exceptions, init_logger, server

logger = init_logger(__name__)


class Proxy(server.Server):
    """Subclass mocking a locally running server. Convenient for testing"""

    def run(self, command, **kwargs):
        """Run command on local server. Exceptions are propagated upwards. Call
        run('stop') as last operation to properly close the databases before
        exiting the process.

        :raises: InvalidRequest on invalid request
                 CommunicationError on unexpected server error
        """
        try:
            response = super().run(command, **kwargs)
        except Exception:
            logger.exception("Unexpected error")
            raise exceptions.CommunicationError("Unexpected error")

        if "error" in response:
            raise exceptions.InvalidRequest(f"Invalid request: {response['error']}")

        return response



import os.path
import tempfile
import unittest

from financeager import (
    DEFAULT_POCKET_NAME,
    DEFAULT_TABLE,
    RECURRENT_TABLE,
    entries,
    exceptions,
    server,
)


class AddEntryToServerTestCase(unittest.TestCase):
    def setUp(self):
        self.server = server.Server()
        self.pocket = "0"
        self.server.run(
            "add",
            name="Hiking boots",
            value=-111.11,
            category="outdoors",
            pocket=self.pocket,
        )

    def test_pocket_name(self):
        self.assertEqual(self.pocket, self.server._pockets[self.pocket].name)

    def test_pockets(self):
        response = self.server.run("pockets")
        self.assertListEqual(response["pockets"], [self.pocket])

    def test_unknown_command(self):
        response = self.server.run("peace")
        self.assertIn("peace", response["error"])

    def test_get_categories(self):
        response = self.server.run("categories", pocket=self.pocket)
        self.assertListEqual(response["categories"], ["outdoors"])


class RecurrentEntryServerTestCase(unittest.TestCase):
    def setUp(self):
        self.server = server.Server()
        self.pocket = "2000"
        self.entry_id = self.server.run(
            "add",
            name="rent",
            value=-1000,
            table_name=RECURRENT_TABLE,
            frequency="monthly",
            start="2000-01-02",
            end="2000-07-01",
            pocket=self.pocket,
        )["id"]

    def test_recurrent_entries(self):
        elements = self.server.run(
            "list", pocket=self.pocket, filters={"name": "rent"}
        )["elements"][RECURRENT_TABLE][self.entry_id]
        self.assertEqual(len(elements), 6)

    def test_recurrent_copy(self):
        destination_pocket = "2001"
        response = self.server.run(
            "copy",
            source_pocket=self.pocket,
            table_name=RECURRENT_TABLE,
            destination_pocket=destination_pocket,
            eid=self.entry_id,
        )
        copied_entry_id = response["id"]
        # copied and added as first element, hence ID 1
        self.assertEqual(copied_entry_id, 1)

        source_entry = self.server.run(
            "get", pocket=self.pocket, table_name=RECURRENT_TABLE, eid=self.entry_id
        )["element"]
        destination_entry = self.server.run(
            "get",
            table_name=RECURRENT_TABLE,
            pocket=destination_pocket,
            eid=copied_entry_id,
        )["element"]
        self.assertDictEqual(source_entry, destination_entry)


class FindEntryServerTestCase(unittest.TestCase):
    def setUp(self):
        self.server = server.Server()
        self.pocket = "0"
        self.entry_id = self.server.run(
            "add", name="Hiking boots", value=-111.11, pocket=self.pocket
        )["id"]

    def test_get_pocket(self):
        pocket = self.server._get_pocket(self.pocket)
        self.assertEqual(pocket.name, self.pocket)

        another_pocket = "foo"
        pocket = self.server._get_pocket(another_pocket)
        self.assertEqual(pocket.name, another_pocket)

        pocket = self.server._get_pocket(None)
        self.assertEqual(pocket.name, DEFAULT_POCKET_NAME)

    def test_get_nonexisting_entry(self):
        response = self.server.run("get", pocket=self.pocket, eid=-1)
        error = response["error"]
        self.assertIsInstance(error, exceptions.PocketEntryNotFound)
        self.assertEqual(str(error), "Entry not found.")

    def test_invalid_update(self):
        response = self.server.run("update", pocket=self.pocket, value="one", eid=1)
        error = response["error"]
        self.assertIsInstance(error, exceptions.PocketValidationFailure)
        self.assertEqual(str(error), "Invalid input data:\nvalue: Not a valid number.")

    def test_query_and_reset_response(self):
        category = entries.CategoryEntry.DEFAULT_NAME
        response = self.server.run(
            "list", pocket=self.pocket, filters={"category": category}
        )
        self.assertGreater(len(response), 0)
        self.assertIsInstance(response, dict)
        self.assertIsInstance(response["elements"], dict)
        self.assertIsInstance(response["elements"][DEFAULT_TABLE], dict)
        self.assertIsInstance(response["elements"][RECURRENT_TABLE], dict)

    def test_response_is_none(self):
        response = self.server.run("get", pocket=self.pocket, eid=self.entry_id)
        self.assertIn("element", response)
        self.assertEqual(response["element"].doc_id, self.entry_id)

        response = self.server.run("remove", pocket=self.pocket, eid=self.entry_id)
        self.assertEqual(response["id"], self.entry_id)

        response = self.server.run(
            "list",
            pocket=self.pocket,
            filters={
                "name": "Hiking boots",
                "category": entries.CategoryEntry.DEFAULT_NAME,
            },
        )
        self.assertDictEqual(response["elements"][DEFAULT_TABLE], {})
        self.assertDictEqual(response["elements"][RECURRENT_TABLE], {})

    def test_update(self):
        new_category = "Outdoorsy shit"
        self.server.run(
            "update", eid=self.entry_id, pocket=self.pocket, category=new_category
        )
        element = self.server.run("get", eid=self.entry_id, pocket=self.pocket)[
            "element"
        ]
        self.assertEqual(element["category"], new_category.lower())

    def test_copy(self):
        destination_pocket = "1"
        response = self.server.run(
            "copy",
            source_pocket=self.pocket,
            destination_pocket=destination_pocket,
            eid=self.entry_id,
        )
        copied_entry_id = response["id"]
        # copied and added as first element, hence ID 1
        self.assertEqual(copied_entry_id, 1)

        source_entry = self.server.run("get", pocket=self.pocket, eid=self.entry_id)[
            "element"
        ]
        destination_entry = self.server.run(
            "get", pocket=destination_pocket, eid=copied_entry_id
        )["element"]
        self.assertDictEqual(source_entry, destination_entry)

    def test_unsuccessful_copy(self):
        self.assertRaises(
            exceptions.PocketEntryNotFound,
            self.server._copy_entry,
            source_pocket=self.pocket,
            destination_pocket="1",
            eid=42,
        )


class JsonPocketsServerTestCase(unittest.TestCase):
    def test_pockets(self):
        tmp_dir = tempfile.mkdtemp()
        self.server = server.Server(data_dir=tmp_dir)

        self.assertDictEqual(self.server.run("pockets"), {"pockets": []})

        # Create dummy JSON files
        pockets = ["1000", "1500"]
        for p in pockets:
            open(os.path.join(tmp_dir, f"{p}.json"), "w").close()

        self.assertDictEqual(self.server.run("pockets"), {"pockets": pockets})


if __name__ == "__main__":
    unittest.main()



"""Data structures for smallest elements of frontend representation of database
query results."""

import time

from . import POCKET_DATE_FORMAT


class Entry:
    """Base class. An entry represents a row in the table that is built from a
    Model.
    The name field is stored in lowercase, simplifying searching from the parent
    listing. The value is rendered absolute to simplify sorting."""

    def __init__(self, name, value):
        """:type name: str
        :type value: float or int
        """
        self.name = name.lower()
        self.value = abs(value)


class BaseEntry(Entry):
    """Innermost element of the Model, child of a CategoryEntry. Holds
    information on name, value, date and eid."""

    DATE_FORMAT = "%y-%m-%d"

    def __init__(self, name, value, date, eid=0):
        """:type eid: int or string, will be converted to int
        :type date: str of valid format
        """
        super().__init__(name, value)
        self.date = time.strftime(
            self.DATE_FORMAT, time.strptime(date, POCKET_DATE_FORMAT)
        )
        self.eid = int(eid)


class CategoryEntry(Entry):
    """First child of the listing, holding BaseEntries. Has a name and a value
    (i.e. the sum of its children's values)."""

    DEFAULT_NAME = "unspecified"

    def __init__(self, name, entries=None):
        """:type entries: list[BaseEntry]"""
        super().__init__(name=name or self.DEFAULT_NAME, value=0.0)

        self.entries = []
        if entries is not None:
            for base_entry in entries:
                self.append(base_entry)

    def append(self, base_entry):
        """Append a BaseEntry to the category and update the value."""
        self.entries.append(base_entry)
        self.value += base_entry.value


def prettify(element, *, default_category):
    """Return element properties formatted as list. The type of the element (recurrent
    or standard) is inferred by the presence of the 'frequency' property.
    If the element's 'category' property is None, use 'default_category'.

    :type element: dict
    :type default_category: str
    """
    recurrent = "frequency" in element

    # Define order of listed properties
    if recurrent:
        properties = ("name", "value", "frequency", "start", "end", "category")
        longest_property_length = 9  # frequency
    else:
        properties = ("name", "value", "date", "category")
        longest_property_length = 8  # category

    if element["category"] is None:
        element["category"] = default_category

    lines = []
    for p in properties:
        lines.append(
            "{}: {}".format(
                p.capitalize().ljust(longest_property_length), str(element[p]).title()
            )
        )

    return "\n".join(lines)



"""Configuration of the financeager application."""

from configparser import ConfigParser, NoOptionError, NoSectionError

from financeager import plugin

from . import init_logger
from .entries import CategoryEntry
from .exceptions import InvalidConfigError

logger = init_logger(__name__)


class Configuration:
    """Wrapper around a ConfigParser object holding configuration. The default
    configuration can be customized by settings specified in a config file.
    """

    def __init__(self, filepath=None, plugins=None):
        """Initialize the default configuration, overwrite with custom
        configuration from file if available, and eventually validate the loaded
        configuration.
        If plugins are given, their configuration is taken into account.

        :type plugins: list[plugin.PluginBase]

        :raises: InvalidConfigError if validation fails
        """
        self._filepath = filepath

        self._plugins = plugins or []
        self._parser = ConfigParser()
        self._option_types = {}

        self._init_defaults()
        self._init_option_types()

        self._load_custom_config()
        self._validate()

    def _init_defaults(self):
        self._parser["SERVICE"] = {
            "name": "local",
        }
        self._parser["FRONTEND"] = {
            "default_category": CategoryEntry.DEFAULT_NAME,
        }

        for p in self._plugins:
            p.config.init_defaults(self._parser)

    def _init_option_types(self):
        for p in self._plugins:
            p.config.init_option_types(self._option_types)

    def _load_custom_config(self):
        """Update config values according to customization in config file."""
        if self._filepath is None:
            return

        logger.debug(f"Loading custom config from {self._filepath}")

        custom_config = ConfigParser()
        # read() silently ignores non-existing paths but returns list of paths
        # that were successfully read
        used_filepaths = custom_config.read(self._filepath)
        if len(used_filepaths) < 1 or used_filepaths[0] != self._filepath:
            raise InvalidConfigError("Config filepath does not exist!")

        for section in self._parser.sections():
            for item in self._parser.options(section):
                try:
                    # read raw value to avoid issues when reading date format
                    # containing percent-characters
                    custom_value = custom_config.get(section, item, raw=True)
                except (NoSectionError, NoOptionError):
                    # section or option not specified by user
                    continue
                self._parser[section][item] = custom_value

    def get_section(self, section):
        """Return a dictionary of options of the requested section.
        If an option is typed, a converted value is returned.
        """
        return {o: self.get_option(section, o) for o in self._parser.options(section)}

    def get_option(self, section, option):
        """Return the requested option of the configuration.
        If an option is typed, a converted value is returned.
        """
        try:
            option_type = self._option_types[section][option]
        except KeyError:
            # Option type not specified, assuming str
            option_type = None

        if option_type in ("int", "float", "boolean"):
            get = getattr(self._parser, f"get{option_type}")
        else:
            get = self._parser.get

        return get(section, option)

    def _validate(self):
        """Validate certain options of the configuration. Typed options are
        validated for possible conversion.

        :raises: InvalidConfigError
        """
        valid_services = ["local"]
        for p in self._plugins:
            if isinstance(p, plugin.ServicePlugin):
                valid_services.append(p.name)

        service_name = self.get_option("SERVICE", "name")
        if service_name not in valid_services:
            raise InvalidConfigError(f"Unknown service name: {service_name}")

        if len(self.get_option("FRONTEND", "default_category")) < 1:
            raise InvalidConfigError("Default category name too short!")

        for section in self._option_types:
            for option in self._option_types[section]:
                try:
                    self.get_option(section, option)
                except ValueError:
                    raise InvalidConfigError(
                        f"Wrong type for option {option} in section {section}."
                    )

        for p in self._plugins:
            p.config.validate(self)



"""Infrastructure for backend communication."""

import os.path
import traceback
from collections import namedtuple

import financeager

from . import exceptions, init_logger, localserver, plugin

logger = init_logger(__name__)


def create(*, configuration, sinks, plugins):
    """Factory to create the Client subclass suitable to the given
    configuration.
    Clients of service plugins are taken into account if specified.
    The sinks are passed into the Client.
    """
    clients = {
        "local": LocalServerClient,
    }

    plugins = plugins or []
    for p in plugins:
        if isinstance(p, plugin.ServicePlugin):
            clients[p.name] = p.client

    service_name = configuration.get_option("SERVICE", "name")
    client_class = clients[service_name]

    return client_class(configuration=configuration, sinks=sinks)


class Client:
    """Abstract interface for communicating with the service.
    Output is directed to distinct sinks which are functions taking a single
    string argument.
    """

    Sinks = namedtuple("Sinks", ["info", "error"])

    def __init__(self, *, configuration, sinks):
        """Store the specified configuration and sinks as attributes.
        The subclass implementation must set up the proxy.
        """
        self.proxy = None
        self.configuration = configuration
        self.sinks = sinks
        self.latest_exception = None

    def safely_run(self, command, **params):
        """Execute self.proxy.run() while handling any errors.
        A caught exception is stored in the 'latest_exception' attribute.
        Return whether execution was successful.

        :return: bool
        """
        success = False

        try:
            self.sinks.info(self.proxy.run(command, **params))
            self.latest_exception = None
            success = True
        except exceptions.InvalidRequest as e:
            self.sinks.error(e)
            self.latest_exception = e

        except exceptions.CommunicationError as e:
            self.sinks.error(e)
            self.latest_exception = e

        except Exception as e:
            self.sinks.error(f"Unexpected error: {traceback.format_exc()}")
            self.latest_exception = e

        return success

    def shutdown(self):
        """Routine to run at the end of the Client lifecycle."""


class LocalServerClient(Client):
    """Client for communicating with the financeager localserver."""

    def __init__(self, *, configuration, sinks):
        """Set up proxy."""
        super().__init__(configuration=configuration, sinks=sinks)

        self.proxy = localserver.Proxy(data_dir=financeager.DATA_DIR)

    def safely_run(self, command, **params):
        """Run the parent method, and for certain modifying commands, fetch category
        names from the server and store them in the cache.
        """
        success = super().safely_run(command, **params)
        if command not in ["add", "remove", "update"]:
            return success

        try:
            self._write_categories_for_cli_completion()
        except Exception as e:
            logger.debug(str(e))
        return success

    def _write_categories_for_cli_completion(self):
        # There might be different categories for each pocket. However when reading the
        # cache at the time of building the CLI completion, the target pocket cannot be
        # determined. It's assumed the default pocket is most relevant, hence its
        # contained categories are stored
        categories = self.proxy.run("categories", pocket=None)["categories"]
        fp = os.path.join(financeager.CACHE_DIR, financeager.CATEGORIES_CACHE_FILENAME)
        with open(fp, "w") as f:
            # The category cache is a line-separated list of names
            f.write("\n".join(categories))

    def shutdown(self):
        """Instruct stopping of Server."""
        self.proxy.run("stop")



"""Module defining exceptions of the financeager application."""


class FinanceagerException(Exception):
    """Base class for exceptions raised by financeager."""


class PreprocessingError(FinanceagerException):
    """Error raised during preprocessing of user input."""


class InvalidRequest(FinanceagerException):
    """Server indicated invalid request."""


class CommunicationError(FinanceagerException):
    """Error in client-server communication."""


class InvalidConfigError(FinanceagerException):
    """Invalid configuration encountered."""


class ConversionError(FinanceagerException):
    """Invalid input for database conversion."""


class PocketException(FinanceagerException):
    """Base exception for pockets module."""


class PocketValidationFailure(PocketException):
    """Invalid input for pocket backend."""


class PocketEntryNotFound(PocketException):
    """Requested entry not found in database."""



import unittest

from financeager import cli, clients, config, plugin

from . import test_config


class TestClient(clients.Client):
    """Dummy class for testing."""


class CreateClientsTestCase(unittest.TestCase):
    def test_create(self):
        # Given the app configuration specifies to use the 'test' service
        app_config = config.Configuration()
        app_config._parser["SERVICE"]["name"] = "test"

        # When the create-factory is invoked
        plugin_config = test_config.TestPluginConfiguration()
        service_plugin = plugin.ServicePlugin(
            config=plugin_config,
            name="test",
            client=TestClient,
        )
        some_plugin = plugin.PluginBase(name="some-plugin", config=plugin_config)
        client = clients.create(
            configuration=app_config,
            sinks=None,
            plugins=[service_plugin, some_plugin],
        )

        # Then the correct client instance is returned
        self.assertIsNone(client.sinks)
        self.assertIsInstance(client, TestClient)

        # And running the CLI parser does not error
        self.assertEqual(
            cli._parse_command(["list"], plugins=[some_plugin])["command"], "list"
        )


if __name__ == "__main__":
    unittest.main()



import shlex
import tempfile
import unittest
from collections import defaultdict
from datetime import datetime as dt
from json import loads as jloads
from unittest import mock

from rich.table import Table as RichTable

import financeager
from financeager import (
    DEFAULT_POCKET_NAME,
    DEFAULT_TABLE,
    RECURRENT_TABLE,
    cli,
    clients,
    config,
    exceptions,
    plugin,
    setup_log_file_handler,
)

TEST_CONFIG_FILEPATH = "/tmp/financeager-test-config"
TEST_DATA_DIR = tempfile.mkdtemp(prefix="financeager-")
setup_log_file_handler(TEST_DATA_DIR)


class CliTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # Create test config file for client
        with open(TEST_CONFIG_FILEPATH, "w") as file:
            file.write(cls.CONFIG_FILE_CONTENT)

        cls.pocket = 1900

    def setUp(self):
        # Separate test runs by running individual test methods using distinct
        # pockets
        self.__class__.pocket += 1

        # Mocks to record output of cli.run() call
        self.info = mock.MagicMock()
        self.error = mock.MagicMock()

    def cli_run(self, command_line, log_method="info", format_args=()):
        """Wrapper around cli.run() function. Adds convenient command line
        options (pocket and config filepath). Executes the actual run() function
        while patching the module logger info and error methods to catch their
        call arguments.

        'command_line' is a string of the form that financeager is called from
        the command line with. 'format_args' are optional objects that are
        formatted into the command line string. Must be passed as tuple if more
        than one.

        If information about an added/update/removed/copied element was to be
        logged, the corresponding ID is matched from the log call arguments to
        the specified 'log_method' and returned. Otherwise the raw log call is
        returned.
        """
        self.info.reset_mock()
        self.error.reset_mock()

        if not isinstance(format_args, tuple):
            format_args = (format_args,)
        args = shlex.split(command_line.format(*format_args))
        command = args[0]

        # Exclude option from subcommand parsers that would be confused
        if command not in ["copy", "pockets"]:
            args.extend(["--pocket", str(self.pocket)])

        args.extend(["--config-filepath", TEST_CONFIG_FILEPATH])

        sinks = clients.Client.Sinks(self.info, self.error)

        # Procedure similar to cli.main()
        params = cli._parse_command(args)
        configuration = config.Configuration(params.pop("config_filepath"))
        exit_code = cli.run(sinks=sinks, configuration=configuration, **params)

        # Get first of the args of the call of specified log method
        response = getattr(self, log_method).call_args[0][0]

        # Verify exit code
        self.assertEqual(
            exit_code, cli.SUCCESS if log_method == "info" else cli.FAILURE
        )

        # Immediately return str messages
        if isinstance(response, str):
            return response

        # Convert Exceptions to string
        if isinstance(response, Exception):
            return str(response)

        if command in ["add", "update", "remove", "copy"]:
            return response["id"]

        if (command in ["get", "pockets"] and log_method == "info") or (
            command == "list" and log_method == "info" and params.get("json")
        ):
            return cli._format_response(
                response,
                command,
                default_category=configuration.get_option(
                    "FRONTEND", "default_category"
                ),
                recurrent_only=params.get("recurrent_only", False),
                entry_sort=params.get("entry_sort"),
                json=params.get("json"),
            )

        if command == "list" and log_method == "info":
            return response

        return response


# Having these app directories patched to None helps testing the error handling for
# category cache reading and writing
@mock.patch("financeager.DATA_DIR", None)
@mock.patch("financeager.CACHE_DIR", None)
class CliLocalServerNoneConfigTestCase(CliTestCase):
    CONFIG_FILE_CONTENT = ""  # service 'local' is the default anyway

    @mock.patch("tinydb.storages.MemoryStorage.write")
    def test_add_entry(self, mocked_write):
        entry_id = str(self.cli_run("add more 1337"))
        # Verify that data is written to memory
        data = mocked_write.call_args[0][0]["standard"][entry_id]
        self.assertEqual(data["name"], "more")
        self.assertEqual(data["value"], 1337.0)

    @mock.patch("builtins.print")
    @mock.patch("financeager.cli.logger.info")
    def test_pockets(self, mocked_info, mocked_print):
        cli.run(
            "pockets", configuration=config.Configuration(filepath=TEST_CONFIG_FILEPATH)
        )

        mocked_info.assert_called_once_with({"pockets": []})
        mocked_print.assert_called_once_with("")

    @mock.patch("builtins.print")
    @mock.patch("financeager.localserver.Proxy.run")
    def test_str_response(self, mocked_run, mocked_print):
        # Cover the behavior of cli._format_response() given a str
        mocked_run.return_value = "response"
        cli.run(
            "pockets", configuration=config.Configuration(filepath=TEST_CONFIG_FILEPATH)
        )

        self.assertListEqual(
            mocked_run.mock_calls, [mock.call("pockets"), mock.call("stop")]
        )
        mocked_print.assert_called_once_with("response")


@mock.patch("financeager.DATA_DIR", TEST_DATA_DIR)
@mock.patch("financeager.CACHE_DIR", TEST_DATA_DIR)
class CliLocalServerTestCase(CliTestCase):
    CONFIG_FILE_CONTENT = """\
[SERVICE]
name = local

[FRONTEND]
default_category = no-category"""

    def test_add_entry(self):
        entry_id = self.cli_run("add entry 42")
        self.assertEqual(entry_id, 1)

        # Verify that customized default category is used
        response = self.cli_run("get {}", format_args=entry_id)
        self.assertIn("no-category", response.lower())

    def test_add_entry_with_date(self):
        entry_id = self.cli_run("add something -1 -d 20-03-04 -c ' very good '")
        self.assertEqual(entry_id, 1)

        response = self.cli_run("get {}", format_args=entry_id)
        self.assertIn("Date    : 2020-03-04", response)
        self.assertIn("Category: Very Good", response)

        self.cli_run("update {} -c -", format_args=entry_id)
        response = self.cli_run("get {}", format_args=entry_id)
        self.assertNotIn("Very Good", response)
        self.assertNotIn("Category: -", response)

    def test_add_entry_with_tablename_and_recurrent(self):
        entry_id = self.cli_run("add stuff 1 -t standard -r")
        self.assertEqual(entry_id, 1)
        response = self.cli_run("get {}", format_args=entry_id)
        self.assertIn("stuff", response.lower())

    def test_add_recurrent_entry_without_tablename(self):
        entry_id = self.cli_run("add flowers -5 -f weekly -e 2020-06-30")
        self.assertEqual(entry_id, 1)

        response = self.cli_run("get {} -t recurrent", format_args=entry_id)
        self.assertIn("Weekly", response)

        self.cli_run("update {} -e -", format_args=entry_id)
        response = self.cli_run("get {} -r", format_args=entry_id)
        self.assertNotIn("2020-06-30", response)

        self.cli_run("update {} -s 2020-01-01", format_args=entry_id)
        response = self.cli_run("get {} -r", format_args=entry_id)
        self.assertIn("2020-01-01", response)

    def test_get_nonexisting_entry(self):
        response = self.cli_run("get 0", log_method="error")
        self.assertEqual(response, "Invalid request: Entry not found.")

    def test_add_empty_name(self):
        response = self.cli_run('add "" 1', log_method="error")
        self.assertEqual(response, "Empty name given.")

    def test_update_empty_category(self):
        response = self.cli_run('update 1 --category ""', log_method="error")
        self.assertEqual(response, "Empty category given.")

    def test_preprocessing_error(self):
        response = self.cli_run("add car -1000 -d ups", log_method="error")
        self.assertEqual(response, "Invalid date format.")

    def test_verbose(self):
        entry_id = self.cli_run("add stuff 100 --verbose")
        response = self.cli_run("get {}", format_args=entry_id)
        self.assertIn("stuff", response.lower())

    def test_list_month(self):
        month_nr = 1
        month_variants = ["01", "Jan", "January"]
        today = dt.today()
        current_month = today.month
        if current_month == month_nr:
            month_nr = 2
            month_variants = ["02", "Feb", "February"]
        month_variants.append(month_nr)
        self.pocket = today.year
        previous_year = self.pocket - 1

        self.cli_run("add beans -4 -d {}-0{}-15", format_args=(self.pocket, month_nr))
        self.cli_run(
            "add chili -4 -d {}-0{}-15", format_args=(self.pocket, month_nr + 1)
        )
        self.cli_run(
            "add tomatoes -6 -d {}-0{}-15", format_args=(previous_year, month_nr)
        )

        for m in month_variants:
            response = self.cli_run("list --month {}", format_args=m)
            names = [v["name"] for v in response["elements"][DEFAULT_TABLE].values()]
            self.assertIn("beans", names)
            self.assertNotIn("chili", names)
            self.assertNotIn("tomatoes", names)

        # Verify overwriting of 'filters' option
        response = self.cli_run(
            "list --month {} --filter date=0{}-", format_args=(month_nr, month_nr + 1)
        )
        names = [v["name"] for v in response["elements"][DEFAULT_TABLE].values()]
        self.assertIn("beans", names)
        self.assertNotIn("chili", names)
        self.assertNotIn("tomatoes", names)

        # Verify default behavior
        response = self.cli_run("list --month")
        names = [v["name"] for v in response["elements"][DEFAULT_TABLE].values()]

        # If it's January: beans -> Feb, chili -> Mar ==> no match
        # Otherwise:       beans -> Jan, chili -> Feb ==> match only in Feb
        if current_month == month_nr + 1:
            self.assertNotIn("beans", names)
            self.assertIn("chili", names)
        else:
            self.assertNotIn("beans", names)
            self.assertNotIn("chili", names)
        self.assertNotIn("tomatoes", names)

    def test_list_invalid_month(self):
        month_nr = 13
        response = self.cli_run("list -m {}", format_args=month_nr, log_method="error")
        self.assertEqual(response, f"Invalid month: {month_nr}")

    def test_list_filter_value(self):
        entry_id = self.cli_run("add money 10")
        response = self.cli_run("list -f value=10")
        element = response["elements"][DEFAULT_TABLE][entry_id]
        self.assertEqual(element["name"], "money")
        self.assertEqual(element["value"], 10.0)
        response = self.cli_run("list -f value=10 -f name=cash")["elements"]
        self.assertEqual(
            response, {DEFAULT_TABLE: {}, RECURRENT_TABLE: defaultdict(list)}
        )

    def test_list_recurrent_only(self):
        id1 = self.cli_run("add interest 20 -s 2020-01-01 -f yearly -c banking")
        id2 = self.cli_run("add rent -300 -s 2020-06-15 -e 2021-12-31 -f monthly")
        response = self.cli_run("list --recurrent-only")
        self.assertEqual(
            response,
            {
                "elements": [
                    {
                        "eid": id1,
                        "name": "interest",
                        "value": 20.0,
                        "start": "2020-01-01",
                        "end": None,
                        "frequency": "yearly",
                        "category": "banking",
                    },
                    {
                        "eid": id2,
                        "name": "rent",
                        "value": -300.0,
                        "start": "2020-06-15",
                        "end": "2021-12-31",
                        "frequency": "monthly",
                        "category": None,
                    },
                ]
            },
        )

        response = self.cli_run("list --recurrent-only -f name=int")
        self.assertEqual(
            response,
            {
                "elements": [
                    {
                        "eid": id1,
                        "name": "interest",
                        "value": 20.0,
                        "start": "2020-01-01",
                        "end": None,
                        "frequency": "yearly",
                        "category": "banking",
                    },
                ]
            },
        )

        response = self.cli_run("list --recurrent-only -f frequency=month")
        self.assertEqual(
            response,
            {
                "elements": [
                    {
                        "eid": id2,
                        "name": "rent",
                        "value": -300.0,
                        "start": "2020-06-15",
                        "end": "2021-12-31",
                        "frequency": "monthly",
                        "category": None,
                    },
                ]
            },
        )

    def test_list_json(self):
        entry_id = self.cli_run("add money 10")
        response = self.cli_run("list --json")
        today = dt.today().strftime("%Y-%m-%d")
        self.assertEqual(
            jloads(response),
            {
                DEFAULT_TABLE: {
                    str(entry_id): {
                        "category": None,
                        "date": today,
                        "name": "money",
                        "value": 10.0,
                    }
                },
                RECURRENT_TABLE: {},
            },
        )

    @mock.patch("financeager.server.Server.run")
    def test_communication_error(self, mocked_run):
        # Raise exception on first call, behave fine on stop call
        mocked_run.side_effect = [Exception(), {}]
        response = self.cli_run("list", log_method="error")
        self.assertEqual(response, "Unexpected error")

    @mock.patch("financeager.localserver.Proxy.run")
    def test_unexpected_error(self, mocked_run):
        # Raise exception on first call, behave fine on stop call
        mocked_run.side_effect = [Exception(), {}]
        response = self.cli_run("list", log_method="error")
        self.assertTrue(response.startswith("Unexpected error: Traceback"))

    def test_add_recurrent_entry(self):
        entry_id = self.cli_run("add credit -100 --recurrent -f monthly -s 01-01")
        self.assertEqual(entry_id, 1)

        year = dt.today().year
        response = self.cli_run("get {} -r", format_args=entry_id)
        self.assertEqual(
            response,
            f"""Name     : Credit
Value    : -100.0
Frequency: Monthly
Start    : {year}-01-01
End      : None
Category : No-Category""",
        )

        entry_id = self.cli_run("add gift -200 -t recurrent -f quarter-yearly -e 12-31")
        self.assertEqual(entry_id, 2)

        response = self.cli_run("get {} -t recurrent", format_args=entry_id)

        self.assertIn(f"End      : {year}-12-31", response)


@mock.patch("financeager.DATA_DIR", TEST_DATA_DIR)
@mock.patch("financeager.CACHE_DIR", TEST_DATA_DIR)
class CategoryCacheTestCase(CliTestCase):
    CONFIG_FILE_CONTENT = ""  # service 'local' is the default anyway

    def setUp(self):
        super().setUp()
        self.pocket = DEFAULT_POCKET_NAME

    def test_cli_categories_cache(self):
        # Initially, the cache is empty
        categories = sorted(cli._read_categories_for_cli_completion())
        self.assertListEqual(categories, [])

        # Adding entries adds their categories to the cache without duplication
        entry_ids = []
        for category in "abcdecd":
            entry_id = self.cli_run(f"add something -10 -c {category}")
            entry_ids.append(entry_id)
        categories = sorted(cli._read_categories_for_cli_completion())
        self.assertListEqual(categories, list("abcde"))

        # Remove entries with category "d"
        self.cli_run(f"remove {entry_ids[-1]}")
        categories = sorted(cli._read_categories_for_cli_completion())
        self.assertListEqual(categories, list("abcde"))
        self.cli_run(f"remove {entry_ids[-4]}")
        categories = sorted(cli._read_categories_for_cli_completion())
        self.assertListEqual(categories, list("abce"))

        # Update entry with category "a" to category "f"
        self.cli_run(f"update {entry_ids[0]} -c f")
        categories = sorted(cli._read_categories_for_cli_completion())
        self.assertListEqual(categories, list("bcef"))


class PreprocessTestCase(unittest.TestCase):
    @unittest.skip("DD.MM. not recognized as date format by dateutil")
    def test_date_format(self):
        data = {"date": "31.01."}
        cli._preprocess(data)
        self.assertDictEqual(data, {"date": "01-31"})

    def test_date(self):
        data = {"date": "02-28"}
        cli._preprocess(data)
        self.assertDictEqual(data, {"date": f"{dt.today().year}-02-28"})

    def test_date_format_error(self):
        data = {"date": "01_01"}
        self.assertRaises(
            exceptions.PreprocessingError,
            cli._preprocess,
            data,
        )

    def test_name_filters(self):
        data = {"filters": ["name=italian"]}
        cli._preprocess(data)
        self.assertEqual(data["filters"], {"name": "italian"})

    def test_category_filters(self):
        data = {"filters": ["category=Restaurants"]}
        cli._preprocess(data)
        self.assertEqual(data["filters"], {"category": "restaurants"})

    def test_filters_error(self):
        data = {"filters": ["value-123"]}
        self.assertRaises(exceptions.PreprocessingError, cli._preprocess, data)

    def test_default_category_filter(self):
        data = {"filters": ["category=unspecified"]}
        cli._preprocess(data)
        self.assertEqual(data["filters"], {"category": None})

    def test_month_filter(self):
        data = {"month": "Jan"}
        cli._preprocess(data)
        self.assertEqual(data["filters"], {"date": f"{dt.today().year}-01-"})

    def test_recurrent_only_fields_filter(self):
        filters = {
            "frequency": "year",
            "start": "2021-",
            "end": "2020-",
        }
        for name, value in filters.items():
            data = {"filters": [f"{name}={value}"]}
            cli._preprocess(data)
            self.assertEqual(data["filters"], {name: value})
            self.assertTrue(data["recurrent_only"])

    def test_filter_indefinite_end(self):
        data = {"filters": ["end="]}
        cli._preprocess(data)
        self.assertEqual(data["filters"], {"end": None})


class FormatResponseTestCase(unittest.TestCase):
    def test_add(self):
        self.assertEqual("Added element 1.", cli._format_response({"id": 1}, "add"))

    def test_update(self):
        self.assertEqual(
            "Updated element 1.", cli._format_response({"id": 1}, "update")
        )

    def test_remove(self):
        self.assertEqual(
            "Removed element 1.", cli._format_response({"id": 1}, "remove")
        )

    def test_copy(self):
        self.assertEqual("Copied element 1.", cli._format_response({"id": 1}, "copy"))

    def test_list(self):
        self.assertEqual(
            "No entries found.",
            cli._format_response(
                {"elements": {DEFAULT_TABLE: {}, RECURRENT_TABLE: {}}}, "list"
            ),
        )

        for options in [{}, {"stacked_layout": True}, {"category_percentage": True}]:
            self.assertIsInstance(
                cli._format_response(
                    {
                        "elements": {
                            DEFAULT_TABLE: {
                                1: {
                                    "name": "icecream",
                                    "value": -5,
                                    "date": "2020-06-25",
                                    "category": "food",
                                },
                            },
                            RECURRENT_TABLE: {},
                        }
                    },
                    "list",
                    **options,
                ),
                RichTable,
            )

    def test_list_recurrent_only(self):
        self.assertIsInstance(
            cli._format_response({"elements": []}, "list", recurrent_only=True),
            RichTable,
        )
        self.assertIsInstance(
            cli._format_response(
                {
                    "elements": [
                        {
                            "eid": 1,
                            "name": "rent",
                            "value": -500,
                            "start": "2020-01-01",
                            "end": None,
                            "frequency": "monthly",
                            "category": "living",
                        }
                    ]
                },
                "list",
                recurrent_only=True,
            ),
            RichTable,
        )


class TestPluginCliOptions(plugin.PluginCliOptions):
    def extend(self, command_parser):
        bird_parser = command_parser.add_parser("bird")
        bird_parser.add_argument("--sound")


class TestClient(clients.Client):
    def safely_run(self, command, **params):
        if command != "bird":
            raise ValueError

        if "sound" not in params:
            raise ValueError

        return True


class PluginCliOptionsTestCase(unittest.TestCase):
    def test_cli_options(self):
        test_plugin = plugin.ServicePlugin(
            name="test-plugin",
            config=None,
            cli_options=TestPluginCliOptions(),
            client=TestClient,
        )

        args = cli._parse_command("bird --sound tweet".split(), plugins=[test_plugin])
        self.assertEqual(args["command"], "bird")
        self.assertEqual(args["sound"], "tweet")

        configuration = config.Configuration()
        configuration._parser["SERVICE"] = {"name": "test-plugin"}
        exit_code = cli.run(**args, configuration=configuration, plugins=[test_plugin])
        self.assertEqual(exit_code, cli.SUCCESS)


class AppDirectoryTestCase(unittest.TestCase):
    def test_dirs(self):
        self.assertTrue(financeager.CONFIG_DIR.endswith(".config/financeager"))
        self.assertTrue(financeager.DATA_DIR.endswith(".local/share/financeager"))
        self.assertTrue(financeager.LOG_DIR.endswith(".cache/financeager/log"))
        self.assertTrue(financeager.CACHE_DIR.endswith(".cache/financeager"))


if __name__ == "__main__":
    unittest.main()



import tempfile
import time
import unittest

from financeager import plugin
from financeager.config import Configuration, InvalidConfigError


class ConfigTestCase(unittest.TestCase):
    def test_sections(self):
        config = Configuration()
        self.assertSetEqual(set(config._parser.sections()), {"SERVICE", "FRONTEND"})

    def test_get_option(self):
        config = Configuration()
        self.assertEqual(config.get_option("SERVICE", "name"), "local")
        self.assertDictEqual(config.get_section("SERVICE"), {"name": "local"})

    def test_invalid_config(self):
        filepath = f"/tmp/{int(time.time())}"

        for content in (
            "[SERVICE]\nname = sillyservice\n",
            "[FRONTEND]\ndefault_category = ",
        ):
            with open(filepath, "w") as file:
                file.write(content)
            self.assertRaises(InvalidConfigError, Configuration, filepath=filepath)

    def test_nonexisting_config_filepath(self):
        filepath = f"/tmp/{time.time()}"
        with self.assertRaises(InvalidConfigError) as cm:
            Configuration(filepath=filepath)
        self.assertTrue(
            cm.exception.args[0].endswith("Config filepath does not exist!")
        )


class TestPluginConfiguration(plugin.PluginConfiguration):
    def init_defaults(self, config_parser):
        config_parser["TESTSECTION"] = {"test": 42, "flag": "off", "number": 1.0}

    def init_option_types(self, option_types):
        option_types["TESTSECTION"] = {
            "test": "int",
            "flag": "boolean",
            "number": "float",
        }


class PluginConfigTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.plugin = plugin.PluginBase(
            name="test-plugin", config=TestPluginConfiguration()
        )

    def test_init_defaults(self):
        config = Configuration(plugins=[self.plugin])

        # Since creating the Configuration instance did not fail, validation was
        # successful
        self.assertEqual(config.get_option("TESTSECTION", "test"), 42)
        self.assertFalse(config.get_option("TESTSECTION", "flag"))
        self.assertEqual(config.get_option("TESTSECTION", "number"), 1.0)

        self.assertDictEqual(
            config.get_section("TESTSECTION"),
            {"test": 42, "flag": False, "number": 1.0},
        )

    def test_load_custom_test_section(self):
        filepath = tempfile.mkstemp()[1]
        with open(filepath, "w") as file:
            file.write("[TESTSECTION]\ntest = 84\n")

        config = Configuration(filepath=filepath, plugins=[self.plugin])
        self.assertEqual(config.get_option("TESTSECTION", "test"), 84)

    def test_validate(self):
        filepath = tempfile.mkstemp()[1]
        with open(filepath, "w") as file:
            file.write("[TESTSECTION]\ntest = no-int\n")

        self.assertRaises(
            InvalidConfigError, Configuration, filepath=filepath, plugins=[self.plugin]
        )

    def test_load_service_plugin_config(self):
        pl = plugin.ServicePlugin(
            name="test-plugin", config=TestPluginConfiguration(), client=None
        )
        filepath = tempfile.mkstemp()[1]
        with open(filepath, "w") as file:
            file.write(f"[SERVICE]\nname = {pl.name}\n")

        config = Configuration(filepath=filepath, plugins=[pl])
        self.assertEqual(config.get_option("SERVICE", "name"), pl.name)


if __name__ == "__main__":
    unittest.main()



"""Support for plugin development."""

import abc


class PluginConfiguration(abc.ABC):
    """Utility to provide configuration specific to the plugin.
    When developing a plugin, provide an implementation derived from this
    class."""

    @abc.abstractmethod
    def init_defaults(self, config_parser):
        """Initialize configuration defaults by extending the given ConfigParser
        by relevant sections.
        """

    def init_option_types(self, option_types):
        """Specify types of plugin options (int, float, boolean) if conversion
        at the time of retrieving the option is desired. The given dictionary
        should be modified in-place by adding the type to the corresponding
        section and option.
        """

    def validate(self, config):
        """Validate content of the Configuration object specific to the plugin.
        Typed options are implicitly validated for conversion by the
        Configuration object already.
        """


class PluginCliOptions(abc.ABC):
    """Utility to provide CLI options specific to the plugin.
    When developing a plugin, optionally provide an implementation derived from
    this class."""

    @abc.abstractmethod
    def extend(self, command_parser):
        """Extend the 'command' subparser of the financeager CLI by additional commands
        and/or arguments.
        Remember that the corresponding Client class must provide an
        implementation to handle the new commands.
        """


class DefaultPluginCliOptions(PluginCliOptions):
    def extend(self, _):
        pass


class PluginBase:
    def __init__(self, *, name, config, cli_options=None):
        """Set plugin name, config (a PluginConfiguration instance), and optional CLI
        options (a PluginCliOptions instance)."""
        self.name = name
        self.config = config
        self.cli_options = cli_options or DefaultPluginCliOptions()


class ServicePlugin(PluginBase):
    """Container for a financeager.services plugin.
    When developing a service plugin, provide a Client class required for
    communication with the service.
    """

    def __init__(self, *, name, config, client, cli_options=None):
        super().__init__(name=name, config=config, cli_options=cli_options)
        self.client = client



"""Tabular, frontend-representation of financeager pocket."""

from json import dumps as jdumps

from . import DEFAULT_TABLE, RECURRENT_TABLE
from .entries import BaseEntry, CategoryEntry
from .rich import richify_listings, richify_recurrent_elements


class Listing:
    """Holds Entries in hierarchical order. First-level children are
    CategoryEntries, second-level children are BaseEntries. Generator methods
    are provided to iterate over these."""

    def __init__(self, name=None, categories=None):
        self.name = name or "Listing"
        self.categories = categories or []

    @classmethod
    def from_elements(cls, elements, default_category=None, name=None):
        """Create listing from list of element dictionaries"""
        listing = cls(name=name)
        for element in elements:
            category = element.pop("category", None) or default_category
            listing.add_entry(BaseEntry(**element), category_name=category)
        return listing

    def add_entry(self, entry, category_name=None):
        """Add a Category- or BaseEntry to the listing.
        Category names are unique, i.e. a CategoryEntry is discarded if one
        with identical name (case INsensitive) already exists.
        When adding a BaseEntry, the parent CategoryEntry is created if it does
        not exist. If no category is specified, the BaseEntry is added to the
        default category.

        :raises: TypeError if neither CategoryEntry nor BaseEntry given
        """
        if isinstance(entry, CategoryEntry):
            if entry.name not in self.category_entry_names:
                self.categories.append(entry)
        elif isinstance(entry, BaseEntry):
            category_entry = self._get_category_entry(category_name)
            category_entry.append(entry)
        else:
            raise TypeError(f"Invalid entry type: {entry}")

    def category_fields(self, field_type):
        """Generator iterating over the field specified by `field_type` of the
        first-level children (CategoryEntries) of the listing.

        :param field_type: 'name' or 'value'

        raises: KeyError if `field_type` not found.
        yields: str or float
        """
        for category_entry in self.categories:
            yield getattr(category_entry, field_type)

    @property
    def category_entry_names(self):
        """Convenience generator method yielding category names."""
        for category_name in self.category_fields("name"):
            yield category_name

    def _get_category_entry(self, category_name):
        """Fetch CategoryEntry searching for given `category_name` or return a
        new instance that is automatically added to the Listing's categories.
        The search is case insensitive.

        :return: CategoryEntry
        """
        category_name = category_name.lower()

        for category_entry in self.categories:
            if category_entry.name == category_name:
                return category_entry
        else:
            # Nothing found in existing categories
            category_entry = CategoryEntry(name=category_name)
            self.add_entry(category_entry)
            return category_entry

    def total_value(self):
        """Return total value of the listing."""
        return sum(v for v in self.category_fields("value"))


def prettify(
    elements,
    recurrent_only=False,
    json=False,
    default_category=None,
    **listing_options,
):
    """Sort the given elements (type acc. to Pocket._search_all_tables) by
    positive and negative value and print tabular representation.

    :param json: If True, return elements as JSON-formatted string
    :param recurrent_only: If True, assume that given elements are purely
        recurrent ones
    :param listing_options: Options passed to rich.richify_listings()
    """
    if json:
        return jdumps(elements)

    if recurrent_only:
        entry_sort = listing_options.get("entry_sort")
        return richify_recurrent_elements(elements, entry_sort=entry_sort)

    listings = _derive_listings(elements, default_category=default_category)
    return richify_listings(listings, **listing_options)


def _derive_listings(elements, *, default_category):
    earnings = []
    expenses = []

    def _sort(eid, element):
        # Copying avoids modifying the original element. Flattening is in order
        # to distinguish recurrent entries (they have the same element ID which
        # thus can't be used as dict key)
        flat_element = element.copy()
        flat_element["eid"] = eid
        if flat_element["value"] > 0:
            earnings.append(flat_element)
        else:
            expenses.append(flat_element)

    # process standard elements
    for eid, element in elements[DEFAULT_TABLE].items():
        _sort(eid, element)

    # process recurrent elements, i.e. for each eid iterate list
    for eid, recurrent_elements in elements[RECURRENT_TABLE].items():
        for element in recurrent_elements:
            _sort(eid, element)

    if not earnings and not expenses:
        return

    listing_earnings = Listing.from_elements(
        earnings, default_category=default_category, name="Earnings"
    )
    listing_expenses = Listing.from_elements(
        expenses, default_category=default_category, name="Expenses"
    )
    return [listing_earnings, listing_expenses]



"""Defines Pocket database object holding per-year financial data."""

import os.path
from abc import ABC, abstractmethod
from collections import Counter, defaultdict
from datetime import datetime as dt

from dateutil import rrule
from marshmallow import Schema, ValidationError, fields, validate
from tinydb import Query, TinyDB, storages
from tinydb.database import Document

from . import (
    DEFAULT_POCKET_NAME,
    DEFAULT_TABLE,
    POCKET_DATE_FORMAT,
    RECURRENT_TABLE,
    UNSET_INDICATOR,
    exceptions,
)

_DEFAULT_CATEGORY = None
FREQUENCY_CHOICES = [
    "yearly",
    "half-yearly",
    "quarter-yearly",
    "bimonthly",
    "monthly",
    "weekly",
    "daily",
]


class EntryBaseSchema(Schema):
    name = fields.String(
        required=True, validate=validate.Length(min=1), allow_none=True
    )
    value = fields.Float(required=True, allow_none=True)
    category = fields.String(validate=validate.Length(min=1), load_default=None)


class StandardEntrySchema(EntryBaseSchema):
    date = fields.Date(format=POCKET_DATE_FORMAT, load_default=None)


class RecurrentEntrySchema(EntryBaseSchema):
    frequency = fields.String(
        validate=validate.OneOf(choices=FREQUENCY_CHOICES),
        required=True,
        allow_none=True,
    )
    start = fields.Date(format=POCKET_DATE_FORMAT, load_default=None)
    end = fields.Date(format=POCKET_DATE_FORMAT, load_default=None)


class Pocket(ABC):
    def __init__(self, name=None):
        """Create Pocket object. Its name defaults to the current year if not
        specified.
        """
        self._name = f"{name or DEFAULT_POCKET_NAME}"

    @property
    def name(self):
        return self._name

    @abstractmethod
    def add_entry(self, table_name=None, **kwargs):
        """Add an entry (standard or recurrent) to the database.

        If 'table_name' is not specified, the kwargs name, value[, category,
        date] are used to insert a unique entry in the standard table.
        With 'table_name' as 'recurrent', the kwargs name, value, frequency
        [, start, end, category] are used to insert a template entry in the
        recurrent table.

        Two kwargs are mandatory:
            :param name: entry name
            :type name: str
            :param value: entry value
            :type value: float, int or str

        The following kwarg is optional:
            :param category: entry category. If not specified, the program
                attempts to derive it from previous, eponymous entries. If this
                fails, the default category is assigned
            :type category: str or None

        The following kwarg is optional for standard entries:
            :param date: entry date. Defaults to current date
            :type date: str of date format

        The following kwarg is mandatory for recurrent entries:
            :param frequency: 'yearly', 'half-yearly', 'quarter-yearly',
                'bimonthly', 'monthly', 'weekly' or 'daily'

        The following kwargs are optional for recurrent entries:
            :param start: start date (defaults to current date)
            :param end: end date (defaults to None which evaluates to the
                current day when the entry is queried)

        :raise: PocketValidationFailure if validation failed
        :return: ID of new entry
        """

    @abstractmethod
    def get_entry(self, eid, table_name=None):
        """Get entry specified by eid in the table table_name.

        :param eid: entry ID
        :type eid: int or str
        :param table_name: table that the entry is stored in (defaults to 'standard')
        :type table_name: str

        :raise: PocketEntryNotFound if element not found
        :return: found element
        """

    @abstractmethod
    def update_entry(self, eid, table_name=None, **kwargs):
        """Update one or more fields of a single entry.

        :param eid: entry ID of the entry to be updated
        :param table_name: table that the entry is stored in (default: 'standard')
        :param kwargs: 'date' for standard entries; any of 'frequency', 'start',
            'end' for recurrent entries; any of 'name', 'value', 'category' for
            either entry type
        :raise: PocketEntryNotFound if element not found
        :return: ID of the updated entry
        """

    @abstractmethod
    def remove_entry(self, eid, table_name=None):
        """Remove an entry from the database given its ID.

        :param eid: ID of the element to be deleted.
        :type eid: int or str
        :param table_name: name of the table that contains the element.
            Default: 'standard'
        :type table_name: str

        :raise: PocketEntryNotFound if element/ID not found.
        :return: element ID if removal was successful
        """

    @abstractmethod
    def get_entries(self, filters=None, recurrent_only=False):
        """Get entries that match the items of the filters dict, if specified.

        If `recurrent_only` is true, return a list of all entries of the
        recurrent table. Filters are applied.

        :param filters: dict of filters to apply
        :param recurrent_only: whether to return only recurrent entries
        :return: entries matching the filters
        """

    @abstractmethod
    def get_categories(self):
        """Return unique category names in alphabetical order."""

    @abstractmethod
    def close(self):
        """Close underlying database."""


class TinyDbPocket(Pocket):
    def __init__(self, name=None, data_dir=None, **kwargs):
        """Create a pocket with a TinyDB database backend, identified by 'name'.
        If 'data_dir' is given, the database storage type is JSON (the storage
        filepath is derived from the Pocket's name). Otherwise the data is
        stored in memory.
        Keyword args are passed to the TinyDB constructor. See the respective
        docs for detailed information.
        """

        super().__init__(name=name)

        # evaluate args/kwargs for TinyDB constructor. This overwrites the
        # 'storage' kwarg if explicitly passed
        if data_dir is None:
            args = []
            kwargs["storage"] = storages.MemoryStorage
        else:
            args = [os.path.join(data_dir, f"{self.name}.json")]
            kwargs["storage"] = storages.JSONStorage

        self._db = TinyDB(*args, **kwargs)
        self._create_category_cache()

    def _create_category_cache(self):
        """The category cache assigns a counter for each element name in the
        database (excluding recurrent elements), keeping track of the
        categories the element was labeled with. This allows deriving the
        category of an element if not explicitly given."""
        self._category_cache = defaultdict(Counter)
        for element in self._db.all():
            self._category_cache[element["name"]].update([element["category"]])

    def _preprocess_entry(self, raw_data=None, table_name=None, partial=False):
        """Perform preprocessing steps (validation, conversion, substitution) of
        raw entry fields prior to adding it to the database.

        :param raw_data: dict containing raw entry fields
        :param table_name: name of the table that the entry is passed to
        :param partial: indicates whether preprocessing is performed before
            adding (False) or updating (True) the database

        :raise: PocketValidationFailure if validation failed or table name
            unknown
        """

        table_name = table_name or DEFAULT_TABLE
        if table_name not in [RECURRENT_TABLE, DEFAULT_TABLE]:
            raise exceptions.PocketValidationFailure(
                f"Unknown table name: {table_name}"
            )

        self._remove_redundant_fields(table_name, raw_data)

        validated_fields = self._validate_entry(
            raw_data=raw_data, table_name=table_name, partial=partial
        )
        converted_fields = self._convert_fields(**validated_fields)

        if not partial:
            converted_fields = self._substitute_none_fields(
                table_name=table_name, **converted_fields
            )

        return converted_fields

    @staticmethod
    def _remove_redundant_fields(table_name, raw_data):
        """The raw data (e.g. parsed from the command line) might contain fields
        that are not required by the given table type and hence, they crash the
        schematics validation ('Rogue field' error). This method removes
        redundant fields in `raw_data` in-place.
        """

        if table_name == RECURRENT_TABLE:
            redundant_fields = ["date"]
        else:
            redundant_fields = ["start", "end", "frequency"]

        for field in redundant_fields:
            raw_data.pop(field, None)

    @staticmethod
    def _validate_entry(raw_data, table_name, **schema_kwargs):
        """Validate raw entry data acc. to ValidationSchema.

        :return: primitive (type-correct) representation of fields
        :raise: PocketValidationFailure if validation failed
        """

        ValidationSchema = (
            RecurrentEntrySchema
            if table_name == RECURRENT_TABLE
            else StandardEntrySchema
        )

        try:
            schema = ValidationSchema(**schema_kwargs)
            validated_data = schema.load(raw_data)
            return schema.dump(validated_data)
        except ValidationError as e:
            infos = [
                f"{field}: {'; '.join(messages)}"
                for field, messages in e.messages.items()
            ]
            raise exceptions.PocketValidationFailure(
                "Invalid input data:\n{}".format("\n".join(infos))
            )

    @staticmethod
    def _convert_fields(**fields):
        """Convert string field values to lowercase for storage. Fields with
        value None are discarded.
        """

        converted_fields = {}

        for k, v in fields.items():
            if v is None:
                continue
            try:
                converted_fields[k] = v.lower()
            except AttributeError:
                converted_fields[k] = v

        return converted_fields

    def _substitute_none_fields(self, table_name, **fields):
        """Substitute optional fields by defaults."""

        substituted_fields = fields.copy()

        # table_name is either of two values; verified in _preprocess_entry
        if table_name == RECURRENT_TABLE:
            if fields.get("start") is None:
                substituted_fields["start"] = dt.today().strftime(POCKET_DATE_FORMAT)
            if fields.get("end") is None:
                substituted_fields["end"] = None
        else:
            if fields.get("date") is None:
                substituted_fields["date"] = dt.today().strftime(POCKET_DATE_FORMAT)

        if fields.get("category") is None:
            name = fields["name"]

            # Find the most common categories previously assigned for the given
            # name. If there is only one OR if there is one that is used more
            # often than any other, assign it; otherwise default to None
            category = _DEFAULT_CATEGORY
            most_common_categories = self._category_cache[name].most_common(2)
            nr_most_common_categories = len(most_common_categories)

            if nr_most_common_categories == 1 or (
                nr_most_common_categories > 1
                and most_common_categories[0][1] != most_common_categories[1][1]
            ):
                category = most_common_categories[0][0]

            substituted_fields["category"] = category

        return substituted_fields

    def _update_category_cache(
        self, eid=None, table_name=None, removing=False, **fields
    ):
        """Update the category cache when adding or updating an entry. The `eid`
        kwarg is used to distinguish the use cases.

        :param eid: element ID when updating
        :param table_name: table name when updating
        :param removing: indicate updating cache after removing an entry
        :param fields: preprossed entry fields to be inserted in the database

        :raise: PocketEntryNotFound if element not found when updating
        """

        if eid is None:
            name = fields["name"]
            category = fields["category"]
            if removing:
                self._category_cache[name][category] -= 1
            else:
                self._category_cache[name].update([category])
        else:
            # raises a PocketEntryNotFound if eid is not found
            old_entry = self.get_entry(eid=eid, table_name=table_name)
            old_name = old_entry["name"]
            old_category = old_entry["category"]

            # update category cache if one of name or category was changed
            if fields.get("name") is not None or fields.get("category") is not None:
                self._category_cache[old_name][old_category] -= 1
                self._category_cache[fields.get("name") or old_name][
                    fields.get("category") or old_category
                ] += 1

    def add_entry(self, table_name=None, **kwargs):
        """Add an entry using TinyDB backend.

        :return: TinyDB ID of new entry (int)
        """

        table_name = table_name or DEFAULT_TABLE
        fields = self._preprocess_entry(raw_data=kwargs, table_name=table_name)

        self._update_category_cache(**fields)

        element_id = self._db.table(table_name).insert(fields)

        return element_id

    def get_entry(self, eid, table_name=None):
        """Get entry using TinyDB backend.

        :return: found element (tinydb.Document)
        """

        table_name = table_name or DEFAULT_TABLE
        element = self._db.table(table_name).get(doc_id=int(eid))
        if element is None:
            raise exceptions.PocketEntryNotFound("Entry not found.")

        return element

    def _preprocess_entry_for_update(self, *, table_name, raw_data):
        """Handle special case for unsetting 'category' and/or 'end' fields while
        preprocessing the entry.
        """
        # Skip field for validation if indicator value present
        end = raw_data.get("end")
        if end == UNSET_INDICATOR:
            del raw_data["end"]
        category = raw_data.get("category")
        if category == UNSET_INDICATOR:
            del raw_data["category"]

        fields = self._preprocess_entry(
            raw_data=raw_data, table_name=table_name, partial=True
        )

        # Unset fields
        if end == UNSET_INDICATOR:
            fields["end"] = None
        if category == UNSET_INDICATOR:
            fields["category"] = None

        return fields

    def update_entry(self, eid, table_name=None, **kwargs):
        """Update entry using TinyDB backend."""

        table_name = table_name or DEFAULT_TABLE
        fields = self._preprocess_entry_for_update(
            raw_data=kwargs, table_name=table_name
        )

        self._update_category_cache(eid=eid, table_name=table_name, **fields)

        element_id = self._db.table(table_name).update(fields, doc_ids=[int(eid)])[0]

        return element_id

    def _search_all_tables(self, condition):
        """Search both the standard table and the recurrent table for elements
        that satisfy the given condition.

        The entries' `doc_id` attribute is used as key in the returned subdicts
        because it is lost in the client-server communication protocol (on
        `financeager print`, the server calls Pocket.get_entries, yet the
        JSON response returned drops the Document.doc_id attribute s.t. it's not
        available when calling prettify on the client side).

        :param condition: condition for the search
        :type condition: tinydb.queries.QueryInstance

        :return: dict
        """

        elements = {DEFAULT_TABLE: {}, RECURRENT_TABLE: defaultdict(list)}

        for element in self._db.search(condition):
            elements[DEFAULT_TABLE][element.doc_id] = element

        # all recurrent elements are generated, and the ones matching the
        # condition are appended to a list that is stored under their generating
        # element's doc_id in the 'recurrent' subdictionary
        for element in self._db.table(RECURRENT_TABLE).all():
            for e in self._create_recurrent_elements(element):
                if condition(e):
                    elements[RECURRENT_TABLE][element.doc_id].append(e)

        return elements

    def _create_recurrent_elements(self, element):
        """Generate elements (holding name, value, category, date) from the
        information of the recurrent element being passed.
        """

        # parse dates to datetime objects
        start = dt.strptime(element["start"], POCKET_DATE_FORMAT)
        now = dt.now()
        end = element["end"]
        end = now if end is None else dt.strptime(end, POCKET_DATE_FORMAT)

        if end > now:
            # don't show entries that are in the future
            end = now

        interval = 1
        frequency = element["frequency"].upper()
        if frequency == "BIMONTHLY":
            frequency = "MONTHLY"
            interval = 2
        elif frequency == "QUARTER-YEARLY":
            frequency = "MONTHLY"
            interval = 3
        elif frequency == "HALF-YEARLY":
            frequency = "MONTHLY"
            interval = 6

        rule = rrule.rrule(
            getattr(rrule, frequency), dtstart=start, until=end, interval=interval
        )

        for date in rule:
            # add date description to name
            name = element["name"]
            if frequency == "YEARLY":
                name = f"{name}, {date.strftime('%Y')}"
            elif frequency == "MONTHLY":
                name = f"{name}, {date.strftime('%B').lower()}"
            elif frequency == "WEEKLY":
                name = f"{name}, week {date.strftime('%W').lower()}"
            else:  # DAILY
                name = f"{name}, day {date.strftime('%-j').lower()}"

            yield Document(
                doc_id=None,
                value=dict(
                    name=name,
                    value=element["value"],
                    category=element["category"],
                    date=date.strftime(POCKET_DATE_FORMAT),
                ),
            )

    def remove_entry(self, eid, table_name=None):
        """Remove entry using TinyDB backend. The category cache is updated."""

        table_name = table_name or DEFAULT_TABLE
        # might raise PocketEntryNotFound if ID not existing
        entry = self.get_entry(eid=int(eid), table_name=table_name)

        self._db.table(table_name).remove(doc_ids=[entry.doc_id])
        self._update_category_cache(removing=True, **entry)

        return entry.doc_id

    @staticmethod
    def _create_query_condition(**filters):
        """Construct query condition according to given filters. A filter is
        given by a key-value pair. The key indicates the field, the value the
        pattern to filter for. Valid keys are 'name', 'date', 'value' and/or
        'category'. Patterns must be of type string, or None (only for the fields
        'category' and 'end; indicates filtering for all entries of the default
        category, and recurrent entries with indefinite end, resp.).
        :return: tinydb.queries.QueryInstance (default: noop)
        """
        condition = Query().noop()
        if not filters:
            return condition

        entry = Query()
        for field, pattern in filters.items():
            if pattern is None and field in ["category", "end"]:
                # The 'category' and 'end' fields are of type string or None. The
                # condition is constructed depending on the filter pattern
                new_condition = entry[field] == None  # noqa
            elif field == "value":
                new_condition = entry[field] == float(pattern)
            else:
                new_condition = entry[field].search(pattern.lower())

            condition &= new_condition

        return condition

    def get_entries(self, filters=None, recurrent_only=False):
        """Get entries using TinyDB backend.

        Constructs a condition from the given filters and uses it to query all tables.

        :return: dict{
                    DEFAULT_TABLE:  dict{ int: tinydb.Document },
                    RECURRENT_TABLE: dict{ int: list[tinydb.Document] }
                    } or
                 list[tinydb.Document]
        """
        filters = filters or {}
        condition = self._create_query_condition(**filters)

        if recurrent_only:
            # Flatten tinydb Document into dict
            return [
                {**e, **{"eid": e.doc_id}}
                for e in self._db.table(RECURRENT_TABLE).search(condition)
            ]

        return self._search_all_tables(condition)

    def get_categories(self):
        """Return unique category names using TinyDB backend."""
        category_names = set(e["category"] for e in self._db.table(DEFAULT_TABLE).all())
        category_names.discard(_DEFAULT_CATEGORY)
        return sorted(category_names)

    def close(self):
        """Close TinyDB database."""
        self._db.close()


TinyDB.default_table_name = DEFAULT_TABLE



import calendar
import datetime as dt
import json
import os.path
import tempfile
import unittest
from collections import Counter

from marshmallow import ValidationError

from financeager import (
    DEFAULT_POCKET_NAME,
    DEFAULT_TABLE,
    POCKET_DATE_FORMAT,
    RECURRENT_TABLE,
    exceptions,
)
from financeager.pocket import (
    _DEFAULT_CATEGORY,
    EntryBaseSchema,
    RecurrentEntrySchema,
    StandardEntrySchema,
    TinyDbPocket,
)


class Entry:
    def __init__(self, **attrs):
        self.__dict__.update(**attrs)


class CreateEmptyPocketTestCase(unittest.TestCase):
    def test_default_name(self):
        pocket = TinyDbPocket()
        self.assertEqual(pocket.name, DEFAULT_POCKET_NAME)
        pocket.close()

    def test_load_category_cache(self):
        # Create data file with one entry and load it into TinyDbPocket
        data_dir = tempfile.mkdtemp(prefix="financeager-")
        name = 1234
        with open(os.path.join(data_dir, f"{name}.json"), "w") as file:
            json.dump(
                {DEFAULT_TABLE: {1: {"name": "climbing", "category": "sport"}}}, file
            )
        pocket = TinyDbPocket(name=name, data_dir=data_dir)
        # Expect that 'sport' has been counted once
        self.assertEqual(pocket._category_cache["climbing"], Counter(["sport"]))
        pocket.close()


class TinyDbPocketStandardEntryTestCase(unittest.TestCase):
    def setUp(self):
        self.pocket = TinyDbPocket(name=1901)
        self.eid = self.pocket.add_entry(
            name="Bicycle", value=-999.99, date="2020-01-01"
        )

    @unittest.skip("missing support for leap year validation in marshmallow")
    def test_leap_year_date(self):
        eid = self.pocket.add_entry(
            name="leap win",
            value=2,
            date="02-29",
        )
        entry = self.pocket.get_entry(eid)
        self.assertEqual(entry["date"], "02-29")

    def test_get_entries(self):
        entries = self.pocket.get_entries(filters={"date": "01-"})
        self.assertEqual("bicycle", entries[DEFAULT_TABLE][1]["name"])

    def test_remove_entry(self):
        response = self.pocket.remove_entry(eid=1)
        self.assertEqual(0, len(self.pocket._db))
        self.assertEqual(1, response)

    def test_create_models_query_kwargs(self):
        eid = self.pocket.add_entry(
            name="Xmas gifts", value=500, date="2002-12-23", category="gifts"
        )
        standard_elements = self.pocket.get_entries(filters={"date": "12"})[
            DEFAULT_TABLE
        ]
        self.assertEqual(len(standard_elements), 1)
        self.assertEqual(standard_elements[eid]["name"], "xmas gifts")

        standard_elements = self.pocket.get_entries(filters={"category": None})[
            DEFAULT_TABLE
        ]
        self.assertEqual(len(standard_elements), 1)

        standard_elements = self.pocket.get_entries(filters={"category": "gi"})[
            DEFAULT_TABLE
        ]
        self.assertEqual(len(standard_elements), 1)

        standard_elements = self.pocket.get_entries(filters={"value": 500})[
            DEFAULT_TABLE
        ]
        self.assertEqual(len(standard_elements), 1)

        standard_elements = self.pocket.get_entries(
            filters={"category": "gi", "date": "12"}
        )[DEFAULT_TABLE]
        self.assertEqual(len(standard_elements), 1)

        self.pocket.add_entry(name="hammer", value=-33, date="2015-12-20")
        standard_elements = self.pocket.get_entries(
            filters={"name": "xmas", "date": "12"}
        )[DEFAULT_TABLE]
        self.assertEqual(len(standard_elements), 1)
        self.assertEqual(standard_elements[eid]["name"], "xmas gifts")

        standard_elements = self.pocket.get_entries(
            filters={"value": 500, "date": "12"}
        )[DEFAULT_TABLE]
        self.assertEqual(len(standard_elements), 1)

    def test_category_cache(self):
        self.pocket.add_entry(
            name="walmart", value=-50.01, category="groceries", date="1999-02-02"
        )
        self.pocket.add_entry(name="walmart", value=-0.99, date="1999-02-03")

        groceries_elements = self.pocket.get_entries(filters={"category": "groceries"})
        self.assertEqual(len(groceries_elements), 2)
        self.assertEqual(
            sum([e["value"] for e in groceries_elements[DEFAULT_TABLE].values()]), -51
        )

    def test_remove_nonexisting_entry(self):
        self.assertRaises(
            exceptions.PocketEntryNotFound, self.pocket.remove_entry, eid=0
        )

    def test_add_remove_via_eid(self):
        entry_name = "penguin sale"
        entry_id = self.pocket.add_entry(name=entry_name, value=1337, date="2010-12-01")
        nr_entries = len(self.pocket._db)

        removed_entry_id = self.pocket.remove_entry(eid=entry_id)
        self.assertEqual(removed_entry_id, entry_id)
        self.assertEqual(len(self.pocket._db), nr_entries - 1)
        self.assertEqual(self.pocket._category_cache[entry_name][_DEFAULT_CATEGORY], 0)

    def test_get_nonexisting_entry(self):
        self.assertRaises(exceptions.PocketEntryNotFound, self.pocket.get_entry, eid=-1)

    def test_add_entry_default_date(self):
        name = "new backpack"
        entry_id = self.pocket.add_entry(name=name, value=-49.95, date=None)
        element = self.pocket.get_entry(entry_id)
        self.assertEqual(element["date"], dt.date.today().strftime(POCKET_DATE_FORMAT))
        self.pocket.remove_entry(eid=entry_id)

    def test_update_standard_entry(self):
        self.pocket.update_entry(eid=self.eid, value=-100)
        element = self.pocket.get_entry(eid=self.eid)
        self.assertEqual(element["value"], -100)

        # kwargs with None-value should be ignored; they are passed e.g. by the
        # flask_restful RequestParser
        self.pocket.update_entry(
            eid=self.eid, name="Trekking Bicycle", value=None, category=None
        )
        element = self.pocket.get_entry(eid=self.eid)
        self.assertEqual(element["name"], "trekking bicycle")

        self.assertEqual(
            self.pocket._category_cache["bicycle"], Counter({_DEFAULT_CATEGORY: 0})
        )
        self.assertEqual(
            self.pocket._category_cache["trekking bicycle"],
            Counter({_DEFAULT_CATEGORY: 1}),
        )

        self.pocket.update_entry(eid=self.eid, category="Sports", name=None)
        element = self.pocket.get_entry(eid=self.eid)
        self.assertEqual(element["category"], "sports")

        self.assertEqual(
            self.pocket._category_cache["trekking bicycle"],
            Counter({"sports": 1, _DEFAULT_CATEGORY: 0}),
        )

        # string-eids should be internally converted to int
        self.pocket.update_entry(
            eid=str(self.eid), name="MTB Tandem", category="Fun", value=-1000
        )
        element = self.pocket.get_entry(eid=self.eid)
        self.assertEqual(element["name"], "mtb tandem")
        self.assertEqual(element["value"], -1000.0)
        self.assertEqual(element["category"], "fun")

        self.assertEqual(
            self.pocket._category_cache["trekking bicycle"],
            Counter({"sports": 0, _DEFAULT_CATEGORY: 0}),
        )
        self.assertEqual(self.pocket._category_cache["mtb tandem"], Counter({"fun": 1}))

    def test_update_nonexisting_entry(self):
        self.assertRaises(
            exceptions.PocketEntryNotFound,
            self.pocket.update_entry,
            eid=0,
            name="I shall fail",
        )

    def test_add_invalid_entry(self):
        self.assertRaises(
            exceptions.PocketValidationFailure,
            self.pocket.add_entry,
            name="I'm invalid",
            date="1.1",
            value="hundred",
        )

    def test_get_categories(self):
        # Table has one element with default category
        self.assertEqual(len(self.pocket.get_entries()[DEFAULT_TABLE]), 1)
        self.assertEqual(self.pocket.get_categories(), [])
        self.pocket.remove_entry(eid=self.eid)
        self.assertEqual(self.pocket.get_categories(), [])

        category = "Groceries"
        self.pocket.add_entry(name="Apple", value=-5, category=category)
        self.assertEqual(self.pocket.get_categories(), [category.lower()])
        self.pocket.add_entry(name="Pear", value=-3, category=category)
        self.assertEqual(self.pocket.get_categories(), [category.lower()])

        another_category = "Car"
        self.pocket.add_entry(name="Petrol", value=-30, category=another_category)
        self.assertEqual(
            self.pocket.get_categories(), [another_category.lower(), category.lower()]
        )

    def tearDown(self):
        self.pocket.close()


class TinyDbPocketRecurrentEntryNowTestCase(unittest.TestCase):
    def test_no_future_elements_created(self):
        # current year
        pocket = TinyDbPocket()

        elements = pocket.get_entries()
        self.assertEqual(len(elements[DEFAULT_TABLE]), 0)
        self.assertEqual(len(elements[RECURRENT_TABLE]), 0)

        today = dt.date.today().replace(year=1993)
        entry_id = pocket.add_entry(
            table_name=RECURRENT_TABLE,
            name="lunch",
            value=-5,
            frequency="daily",
            start="1993-01-01",
            end=today.strftime(POCKET_DATE_FORMAT),
        )

        day_nr = today.timetuple().tm_yday
        elements = pocket.get_entries()
        self.assertEqual(len(elements[RECURRENT_TABLE][entry_id]), day_nr)


class TinyDbPocketRecurrentEntryTestCase(unittest.TestCase):
    def setUp(self):
        self.pocket = TinyDbPocket(name=1901)

    def test_recurrent_entries(self):
        eid = self.pocket.add_entry(
            name="rent",
            value=-500,
            table_name=RECURRENT_TABLE,
            frequency="monthly",
            start="2007-10-01",
            end="2008-11-30",
        )
        self.assertSetEqual({RECURRENT_TABLE}, self.pocket._db.tables())

        self.assertEqual(len(self.pocket._db.table(RECURRENT_TABLE).all()), 1)
        element = self.pocket._db.table(RECURRENT_TABLE).all()[0]
        recurrent_elements = list(self.pocket._create_recurrent_elements(element))
        self.assertEqual(len(recurrent_elements), 14)

        rep_element_names = {e["name"] for e in recurrent_elements}
        self.assertSetEqual(
            rep_element_names,
            {f"rent, {calendar.month_name[m + 1].lower()}" for m in range(12)},
        )

        matching_elements = self.pocket.get_entries(filters={"date": "11"})[
            RECURRENT_TABLE
        ]
        self.assertEqual(len(matching_elements), 1)
        self.assertEqual(matching_elements[eid][0]["name"], "rent, november")
        # the doc_id attribute is None because a new Document instance has been
        # created in Pocket._create_recurrent_elements. The 'eid' entry
        # however is 1 because the parent element is the first in the
        # recurrent table
        self.assertIsNone(matching_elements[eid][0].doc_id)

    def test_recurrent_quarter_yearly_entries(self):
        eid = self.pocket.add_entry(
            name="interest",
            value=25,
            table_name=RECURRENT_TABLE,
            frequency="quarter-yearly",
            start="1991-01-01",
            end="1991-12-31",
        )

        element = self.pocket._db.table(RECURRENT_TABLE).all()[0]
        recurrent_elements = list(self.pocket._create_recurrent_elements(element))
        self.assertEqual(len(recurrent_elements), 4)

        rep_element_names = {e["name"] for e in recurrent_elements}
        self.assertSetEqual(
            rep_element_names,
            {
                "interest, january",
                "interest, april",
                "interest, july",
                "interest, october",
            },
        )

        recurrent_table_size = len(self.pocket._db.table(RECURRENT_TABLE))
        self.pocket.remove_entry(eid=eid, table_name=RECURRENT_TABLE)
        self.assertEqual(
            len(self.pocket._db.table(RECURRENT_TABLE)), recurrent_table_size - 1
        )

    def test_recurrent_bimonthly_entries(self):
        eid = self.pocket.add_entry(
            name="interest",
            value=25,
            table_name=RECURRENT_TABLE,
            frequency="bimonthly",
            start="2012-01-08",
            end="2012-03-08",
        )
        recurrent_elements = self.pocket.get_entries()[RECURRENT_TABLE][eid]
        self.assertEqual(len(recurrent_elements), 2)
        self.assertEqual(recurrent_elements[0]["name"], "interest, january")

    def test_recurrent_weekly_entries(self):
        eid = self.pocket.add_entry(
            name="interest",
            value=25,
            table_name=RECURRENT_TABLE,
            frequency="weekly",
            start="2000-01-08",
            end="2000-01-14",
        )
        recurrent_elements = self.pocket.get_entries()[RECURRENT_TABLE][eid]
        self.assertEqual(len(recurrent_elements), 1)
        self.assertEqual(recurrent_elements[0]["name"], "interest, week 01")

    def test_get_entries_of_recurrent_table(self):
        interest_fields = dict(
            name="interest",
            value=25.0,
            frequency="weekly",
            start="2000-01-08",
            end="2000-01-14",
        )
        rent_fields = dict(
            name="rent",
            value=-400.0,
            category="living",
            frequency="monthly",
            start="2020-01-01",
        )
        eid = self.pocket.add_entry(table_name=RECURRENT_TABLE, **interest_fields)
        eid2 = self.pocket.add_entry(table_name=RECURRENT_TABLE, **rent_fields)
        recurrent_elements = self.pocket.get_entries(recurrent_only=True)
        interest_fields.update({"category": None, "eid": eid})
        rent_fields.update({"end": None, "eid": eid2})
        self.assertEqual(recurrent_elements, [interest_fields, rent_fields])

    def test_update_recurrent_entry(self):
        eid = self.pocket.add_entry(
            name="interest",
            value=25,
            table_name=RECURRENT_TABLE,
            frequency="quarter-yearly",
            start="2020-01-01",
        )

        self.pocket.update_entry(
            eid=eid,
            frequency="half-yearly",
            start="2020-03-01",
            end="2020-06-30",
            table_name=RECURRENT_TABLE,
        )

        entry = self.pocket.get_entry(eid=eid, table_name=RECURRENT_TABLE)
        self.assertEqual(entry["frequency"], "half-yearly")
        self.assertEqual(entry["start"], "2020-03-01")
        self.assertEqual(entry["end"], "2020-06-30")

        recurrent_entries = self.pocket.get_entries()[RECURRENT_TABLE][eid]
        self.assertEqual(len(recurrent_entries), 1)
        self.assertEqual(recurrent_entries[0]["date"], "2020-03-01")

        self.pocket.update_entry(
            eid=eid, value=30, frequency=None, table_name=RECURRENT_TABLE
        )

    def test_update_recurrent_entry_incorrectly(self):
        eid = self.pocket.add_entry(
            name="interest",
            value=25,
            table_name=RECURRENT_TABLE,
            frequency="quarter-yearly",
            start="2020-01-01",
        )

        with self.assertRaises(exceptions.PocketValidationFailure) as context:
            self.pocket.update_entry(eid=eid, end="Dec-24", table_name=RECURRENT_TABLE)
        self.assertIn("end", str(context.exception))

    def test_recurrent_yearly_entry(self):
        eid = self.pocket.add_entry(
            name="Fee",
            value=-100,
            start="2015-01-01",
            end=None,
            table_name=RECURRENT_TABLE,
            frequency="yearly",
        )
        recurrent_entries = self.pocket.get_entries()[RECURRENT_TABLE][eid]
        self.assertEqual(len(recurrent_entries), dt.date.today().year - 2014)
        self.assertEqual(recurrent_entries[0]["date"], "2015-01-01")
        self.assertEqual(recurrent_entries[0]["name"], "fee, 2015")

    def test_recurrent_entry_ending_in_future(self):
        today = dt.date.today()
        year = today.year
        eid = self.pocket.add_entry(
            name="insurance",
            value=-100,
            start=f"{year}-01-01",
            end=f"{year}-12-31",
            table_name=RECURRENT_TABLE,
            frequency="monthly",
        )
        recurrent_entries = self.pocket.get_entries()[RECURRENT_TABLE][eid]
        self.assertEqual(len(recurrent_entries), today.month)
        self.assertEqual(recurrent_entries[0]["date"], f"{year}-01-01")
        self.assertEqual(recurrent_entries[0]["name"], "insurance, january")

    def test_filter_recurrent_entries_with_indefinite_end(self):
        eid = self.pocket.add_entry(
            name="fees",
            value=-5,
            table_name=RECURRENT_TABLE,
            frequency="monthly",
        )
        recurrent_entries = self.pocket.get_entries(
            filters={"end": None}, recurrent_only=True
        )
        self.assertEqual(
            recurrent_entries,
            [
                {
                    "name": "fees",
                    "value": -5.00,
                    "category": None,
                    "start": dt.date.today().isoformat(),
                    "end": None,
                    "frequency": "monthly",
                    "eid": eid,
                }
            ],
        )

    def tearDown(self):
        self.pocket.close()


class ValidationTestCase(unittest.TestCase):
    def test_valid_base_entry(self):
        data = EntryBaseSchema().load({"name": "entry", "value": "5"})
        entry = Entry(**data)
        self.assertEqual(entry.name, "entry")
        self.assertEqual(entry.value, 5)
        self.assertIsNone(entry.category)

    def test_valid_base_entry_category_none(self):
        data = EntryBaseSchema().load({"name": "entry", "value": "5", "category": None})
        entry = Entry(**data)
        self.assertEqual(entry.name, "entry")
        self.assertEqual(entry.value, 5)
        self.assertIsNone(entry.category)

    def test_valid_standard_entry(self):
        data = StandardEntrySchema().load(
            {"name": "entry", "value": 5, "date": "1900-05-01"}
        )
        entry = Entry(**data)
        self.assertEqual(entry.date, dt.date(year=1900, month=5, day=1))

    def test_valid_standard_entry_default_date(self):
        data = StandardEntrySchema().load({"name": "entry", "value": 5})
        entry = Entry(**data)
        self.assertIsNone(entry.date)

    def test_invalid_base_entry_name(self):
        with self.assertRaises(ValidationError) as context:
            EntryBaseSchema().load({"name": "", "value": 123})
        self.assertListEqual(["name"], list(context.exception.messages.keys()))

    def test_invalid_base_entry_value(self):
        with self.assertRaises(ValidationError) as context:
            EntryBaseSchema().load({"name": "foo", "value": "hundred"})
        self.assertListEqual(["value"], list(context.exception.messages.keys()))

    def test_valid_recurrent_entry(self):
        data = RecurrentEntrySchema().load(
            {
                "name": "rent",
                "value": -400,
                "frequency": "monthly",
                "start": "1900-01-02",
            }
        )
        entry = Entry(**data)
        self.assertEqual(entry.frequency, "monthly")
        self.assertEqual(entry.start, dt.date(year=1900, month=1, day=2))
        self.assertEqual(entry.end, None)

    def test_invalid_recurrent_entry(self):
        with self.assertRaises(ValidationError) as context:
            RecurrentEntrySchema().load(
                {
                    "name": "rent",
                    "value": -400,
                    "frequency": "yaerly",
                    "start": "2020-01-02",
                }
            )
        self.assertListEqual(["frequency"], list(context.exception.messages.keys()))


class ValidateEntryTestCase(unittest.TestCase):
    def setUp(self):
        self.pocket = TinyDbPocket(name=1901)

    def test_validate_valid_standard_entry(self):
        raw_data = {"name": "MoNeY", "value": "124.5"}
        fields = self.pocket._preprocess_entry(raw_data=raw_data)

        self.assertEqual(fields["name"], "money")
        self.assertEqual(fields["value"], 124.5)
        self.assertEqual(len(fields), 4)

    def test_validate_invalid_standard_entry(self):
        raw_data = {"name": "not valid", "value": "hundred"}
        with self.assertRaises(exceptions.PocketValidationFailure) as context:
            self.pocket._preprocess_entry(raw_data=raw_data)
        self.assertIn("value", str(context.exception))

    def test_validate_valid_recurrent_entry(self):
        raw_data = {
            "name": "income",
            "value": "1111",
            "frequency": "bimonthly",
            "start": "2020-06-01",
        }
        fields = self.pocket._preprocess_entry(
            raw_data=raw_data, table_name=RECURRENT_TABLE
        )

        self.assertEqual(fields["frequency"], "bimonthly")
        self.assertEqual(fields["start"], raw_data["start"])
        self.assertEqual(len(fields), 6)

    def test_validate_invalid_recurrent_entry(self):
        raw_data = {
            "name": "income",
            "value": "1111",
            "frequency": "hourly",
            "start": "06-01",
            "category": "",
        }
        with self.assertRaises(exceptions.PocketValidationFailure) as context:
            self.pocket._preprocess_entry(raw_data=raw_data, table_name=RECURRENT_TABLE)
        self.assertIn("frequency", str(context.exception))
        self.assertIn("category", str(context.exception))

    def test_convert_fields(self):
        raw_data = {"name": "CamelCase", "value": 123.0, "category": None}
        converted_fields = self.pocket._convert_fields(**raw_data)

        # None-category is being kicked out
        self.assertEqual(len(raw_data) - 1, len(converted_fields))
        self.assertEqual(converted_fields["name"], "camelcase")
        self.assertEqual(converted_fields["value"], 123.0)

    def test_substitute_none_recurrent_fields(self):
        fields = self.pocket._substitute_none_fields(RECURRENT_TABLE, name="baz")
        self.assertIn("start", fields)
        self.assertIn("end", fields)
        # contains name, category, start, end
        self.assertEqual(len(fields), 4)

    def test_substitute_none_standard_fields(self):
        fields = self.pocket._substitute_none_fields(DEFAULT_TABLE, name="foo")
        self.assertIn("name", fields)
        self.assertIn("date", fields)
        # contains name, date, category
        self.assertEqual(len(fields), 3)

    def test_remove_redundant_fields(self):
        raw_data = {field: None for field in ["date", "start", "end", "frequency"]}
        TinyDbPocket._remove_redundant_fields(None, raw_data)
        self.assertDictEqual(raw_data, {"date": None})
        TinyDbPocket._remove_redundant_fields(RECURRENT_TABLE, raw_data)
        self.assertDictEqual(raw_data, {})

    def test_invalid_table_name(self):
        self.assertRaises(
            exceptions.PocketValidationFailure,
            self.pocket._preprocess_entry,
            table_name="unknown table",
        )


class JsonTinyDbPocketTestCase(unittest.TestCase):
    @classmethod
    def setUp(cls):
        cls.data_dir = "/tmp"
        cls.year = 1999
        cls.pocket = TinyDbPocket(name=cls.year, data_dir=cls.data_dir)
        cls.data_filepath = os.path.join(cls.data_dir, f"{cls.year}.json")

    def test_json_file_exists(self):
        self.assertTrue(os.path.exists(self.data_filepath))

    def test_add_get(self):
        name = "pineapple"
        eid = self.pocket.add_entry(name=name, value=-5)
        element = self.pocket.get_entry(eid=eid)
        self.assertEqual(name, element["name"])
        self.pocket.remove_entry(eid=eid)

    @classmethod
    def tearDown(cls):
        cls.pocket.close()
        os.remove(cls.data_filepath)


if __name__ == "__main__":
    unittest.main()



import unittest

from financeager import DEFAULT_TABLE, RECURRENT_TABLE
from financeager.entries import BaseEntry, CategoryEntry
from financeager.listing import Listing, _derive_listings, prettify


class AddCategoryEntryTestCase(unittest.TestCase):
    def setUp(self):
        self.listing = Listing()
        self.category_name = "Groceries"
        self.listing.add_entry(CategoryEntry(name=self.category_name))

    def test_category_item_in_list(self):
        self.assertIn(self.category_name.lower(), self.listing.category_entry_names)


class AddCategoryEntryTwiceTestCase(unittest.TestCase):
    def setUp(self):
        self.listing = Listing()
        self.category_name = "Groceries"
        self.listing.add_entry(CategoryEntry(name=self.category_name))
        self.listing.add_entry(CategoryEntry(name=self.category_name))

    def test_category_item_in_list(self):
        self.assertIn(self.category_name.lower(), self.listing.category_entry_names)

    def test_single_item_in_list(self):
        self.assertEqual(1, len(list(self.listing.category_entry_names)))


class AddBaseEntryTestCase(unittest.TestCase):
    def setUp(self):
        self.listing = Listing()
        self.item_name = "Aldi"
        self.item_value = 66.6
        self.item_date = "2000-11-08"
        self.item_category = "groceries"
        self.listing.add_entry(
            BaseEntry(self.item_name, self.item_value, self.item_date),
            self.item_category,
        )

    def test_str(self):
        category_entry = self.listing.categories[0]
        self.assertEqual(category_entry.name, self.item_category)
        self.assertEqual(len(category_entry.entries), 1)

    def test_add_invalid_entry(self):
        self.assertRaises(TypeError, self.listing.add_entry, None)


class AddBaseEntryWithoutCategoryTestCase(unittest.TestCase):
    def setUp(self):
        self.listing = Listing()
        self.item_name = "Aldi"
        self.item_value = 66.6
        self.item_date = "2009-11-08"
        self.listing.add_entry(
            BaseEntry(self.item_name, self.item_value, self.item_date),
            category_name=CategoryEntry.DEFAULT_NAME,
        )

    def test_default_category_in_list(self):
        names = list(self.listing.category_entry_names)
        self.assertIn(CategoryEntry.DEFAULT_NAME, names)


class AddTwoBaseEntriesTestCase(unittest.TestCase):
    def setUp(self):
        self.listing = Listing()
        self.item_a_value = 66.6
        self.item_b_value = 10.01
        self.item_category = "Groceries"
        self.date = "2008-11-11"
        self.listing.add_entry(
            BaseEntry("Aldi", self.item_a_value, self.date), self.item_category
        )
        self.listing.add_entry(
            BaseEntry("Rewe", self.item_b_value, self.date), self.item_category
        )

    def test_total_value(self):
        self.assertAlmostEqual(
            self.item_a_value + self.item_b_value, self.listing.total_value(), places=5
        )


class ListingFromElementsTestCase(unittest.TestCase):
    def setUp(self):
        self.name = "Dinner for one"
        self.value = 99.9
        self.date = "1980-12-31"
        self.listing = Listing.from_elements(
            [dict(name=self.name, value=self.value, date=self.date, eid=0)],
            default_category=CategoryEntry.DEFAULT_NAME,
        )

    def test_contains_an_entry(self):
        self.assertEqual(self.date[2:], self.listing.categories[0].entries[0].date)

    def test_category_item_names(self):
        parsed_listing_entry_names = list(self.listing.category_entry_names)
        listing_entry_names = [CategoryEntry.DEFAULT_NAME]
        self.assertListEqual(listing_entry_names, parsed_listing_entry_names)


class PrettifyListingsTestCase(unittest.TestCase):
    def test_prettify_no_elements(self):
        elements = {DEFAULT_TABLE: {}, RECURRENT_TABLE: {}}
        self.assertEqual(prettify(elements), "No entries found.")

    def test_derive_listings(self):
        expense = {
            "name": "food",
            "value": -100.01,
            "date": "2000-03-03",
            "category": "groceries",
        }
        earning = {"name": "money", "value": 299.99, "date": "2000-03-03"}
        recurrent_earning = {
            "name": "gold",
            "value": 4321,
            "date": "2000-01-01",
            "category": "bank",
        }
        elements = {
            DEFAULT_TABLE: {
                1: expense,
                999: earning,
            },
            RECURRENT_TABLE: {42: [recurrent_earning]},
        }
        listing_earnings, listing_expenses = _derive_listings(
            elements, default_category=CategoryEntry.DEFAULT_NAME
        )
        self.assertEqual(listing_earnings.name, "Earnings")
        self.assertEqual(len(listing_earnings.categories), 2)
        for category in listing_earnings.categories:
            self.assertEqual(len(category.entries), 1)
        self.assertEqual(listing_expenses.name, "Expenses")
        self.assertEqual(len(listing_expenses.categories), 1)
        for category in listing_expenses.categories:
            self.assertEqual(len(category.entries), 1)


if __name__ == "__main__":
    unittest.main()



import os.path
from importlib.metadata import version
from logging import DEBUG, WARN, Formatter, StreamHandler, getLogger, handlers

import platformdirs

# versioning information
__version__ = version(__name__)

#
# Global constants
#

# default name of database
DEFAULT_POCKET_NAME = "main"

# fixed date format for database
POCKET_DATE_FORMAT = "%Y-%m-%d"

# indicator char for unsetting field during update
UNSET_INDICATOR = "-"

# tables name of database
DEFAULT_TABLE = "standard"
RECURRENT_TABLE = "recurrent"

# Formatting option defaults
DEFAULT_CATEGORY_ENTRY_SORT_KEY = "value"
DEFAULT_BASE_ENTRY_SORT_KEY = "name"

# directories for application data and log file
CONFIG_DIR = platformdirs.user_config_dir("financeager")
DATA_DIR = platformdirs.user_data_dir("financeager")
CACHE_DIR = platformdirs.user_cache_dir("financeager")
# platformdirs.user_log_dir is ~/.local/state/log which is backwards-incompatible
LOG_DIR = os.path.join(CACHE_DIR, "log")

CONFIG_FILEPATH = os.path.join(CONFIG_DIR, "config")
CATEGORIES_CACHE_FILENAME = "categories"

# Set up the package logger
LOGGER = getLogger(__package__)
LOGGER.setLevel(DEBUG)
_stream_handler = StreamHandler()
_stream_handler.setLevel(WARN)
LOGGER.addHandler(_stream_handler)
FORMATTER = Formatter(fmt="%(levelname)s %(asctime)s %(name)s:%(lineno)d %(message)s")


def init_logger(name):
    """Set up module logger. Library loggers are assigned the package logger as
    parent. Any records are propagated to the parent package logger.
    """
    logger = getLogger(name)
    logger.setLevel(DEBUG)

    if logger.parent.name == "root":
        # Library logger; probably has NullHandler
        logger.parent = LOGGER

    logger.propagate = True

    LOGGER.debug(f"Set up logger: {name}")
    return logger


def setup_log_file_handler(log_dir=LOG_DIR):
    """Create RotatingFileHandler for package logger, storing logs in
    'log_dir' (default: LOG_DIR). The directory is created if not existing.
    """
    os.makedirs(log_dir, exist_ok=True)
    file_handler = handlers.RotatingFileHandler(
        os.path.join(log_dir, "log"), maxBytes=5e6, backupCount=5
    )
    file_handler.setFormatter(FORMATTER)
    LOGGER.addHandler(file_handler)


def make_log_stream_handler_verbose():
    """Make handler show debug messages using more informative format."""
    _stream_handler.setLevel(DEBUG)
    _stream_handler.setFormatter(FORMATTER)



from rich import box
from rich.panel import Panel
from rich.table import Table

from . import DEFAULT_BASE_ENTRY_SORT_KEY, DEFAULT_CATEGORY_ENTRY_SORT_KEY


def richify_listings(
    listings,
    stacked_layout=False,
    category_sort=None,
    category_percentage=False,
    entry_sort=None,
):
    """Create and return rich.Table from listings acc. to given options.
    :param stacked_layout: If True, listings are displayed one by one
    :param category_sort: Field governing category sorting (name, value)
    :param category_percentage: If True, display the share in the listing total of each
        category
    :param entry_sort: Field governing base entry sorting (name, value, date, ID)
    """
    if not listings:
        return "No entries found."

    tables = []
    category_sort = category_sort or DEFAULT_CATEGORY_ENTRY_SORT_KEY
    entry_sort = entry_sort or DEFAULT_BASE_ENTRY_SORT_KEY

    totals = [ls.total_value() for ls in listings]

    # Calculate maximum column widths for stacked layout alignment
    max_widths = [None, None, None, None]
    if stacked_layout:
        max_widths = _calculate_max_column_widths(
            listings, totals, category_percentage, category_sort, entry_sort
        )

    # Build tables for listings
    for listing, total in zip(listings, totals):
        table = Table(
            title=listing.name, show_edge=False, box=box.SIMPLE_HEAVY, expand=True
        )
        table.add_column("Name", min_width=max_widths[0])
        table.add_column("Value", justify="right", min_width=max_widths[1])
        table.add_column(
            "%" if category_percentage else "Date",
            justify="right",
            min_width=max_widths[2],
        )
        table.add_column(
            "" if category_percentage else "ID",
            justify="right",
            min_width=max_widths[3],
        )
        tables.append(table)

        for category in sorted(
            listing.categories, key=lambda e: getattr(e, category_sort)
        ):
            percent = ""
            if category_percentage:
                percent = f"{100 * category.value / total:.1f}"
            table.add_row(
                category.name.title(),
                f"{category.value:0.2f}",
                percent,
                "",
                style="bold",
            )
            if category_percentage:
                continue

            for entry in sorted(category.entries, key=lambda e: getattr(e, entry_sort)):
                # Add two spaces for indent
                table.add_row(
                    f"  {entry.name.title()}",
                    f"{entry.value:.2f}",
                    entry.date,
                    str(entry.eid),
                )

    def nr_rows(listing):
        if category_percentage:
            return len(listing.categories)
        return len(listing.categories) + sum(len(c.entries) for c in listing.categories)

    # Fill shorter table with empty rows
    if not stacked_layout:
        nr_earnings_rows = nr_rows(listings[0])
        nr_expenses_rows = nr_rows(listings[1])
        nr_rows_diff = nr_earnings_rows - nr_expenses_rows
        shorter_table = tables[1] if nr_rows_diff > 0 else tables[0]
        for _ in range(abs(nr_rows_diff)):
            shorter_table.add_row("")

    # Bottom rows for listing totals
    for table, total in zip(tables, totals):
        table.add_row("")
        table.add_row("Total", f"{total:.2f}", "", "", style="bold")

    grid = Table.grid(expand=False, pad_edge=True, padding=(0, 1, 0, 1))
    grid.add_column(min_width=50)

    if stacked_layout:
        # Single column
        grid.add_row(tables[0])
        grid.add_row(tables[1])
    else:
        # Two columns side-by-side
        grid.add_column(min_width=50)
        grid.add_row(*tables)

    diff = totals[0] - totals[1]
    grid.add_row(Panel(f"Difference: {diff:.2f}"), style="red" if diff < 0 else "green")
    return grid


def richify_recurrent_elements(elements, entry_sort=None):
    """Create and return rich.Table from recurrent elements acc. to given options.
    :param entry_sort: Field governing base entry sorting (name, value, ID, category,
        start, end, frequency)
    """
    fields = ["eid", "name", "value", "category", "start", "end", "frequency"]
    table = Table(
        show_edge=False, box=box.SIMPLE_HEAVY, expand=False, row_styles=["i", ""]
    )
    for field in fields:
        field = "id" if field == "eid" else field
        table.add_column(
            field.upper(), justify="right" if field in ["id", "value"] else "left"
        )

    entry_sort = entry_sort or "eid"
    for element in sorted(elements, key=lambda e: e[entry_sort]):
        element["category"] = element["category"] or "Unspecified"
        element["end"] = element["end"] or "-"
        table.add_row(
            *[
                str(element[f]).capitalize() if f != "value" else f"{element[f]:.2f}"
                for f in fields
            ]
        )
    return table


def _calculate_max_column_widths(
    listings, totals, category_percentage, category_sort, entry_sort
):
    """Calculate maximum column widths needed across all listings for
    consistent alignment."""
    max_name_width = 0
    max_value_width = 0
    max_date_percent_width = 0
    max_id_width = 0

    for listing, total in zip(listings, totals):
        # Check category names and values
        for category in sorted(
            listing.categories, key=lambda e: getattr(e, category_sort)
        ):
            # Category name width
            max_name_width = max(max_name_width, len(category.name.title()))

            # Category value width
            value_str = f"{category.value:0.2f}"
            max_value_width = max(max_value_width, len(value_str))

            # Date/percentage column width for categories
            if category_percentage:
                percent_str = f"{100 * category.value / total:.1f}"
                max_date_percent_width = max(max_date_percent_width, len(percent_str))

            if not category_percentage:
                # Check entry names, values, dates, and IDs
                for entry in sorted(
                    category.entries, key=lambda e: getattr(e, entry_sort)
                ):
                    # Entry name width (with 2-space indent)
                    entry_name = f"  {entry.name.title()}"
                    max_name_width = max(max_name_width, len(entry_name))

                    # Entry value width
                    entry_value_str = f"{entry.value:.2f}"
                    max_value_width = max(max_value_width, len(entry_value_str))

                    # Entry date width
                    max_date_percent_width = max(
                        max_date_percent_width, len(entry.date)
                    )

                    # Entry ID width
                    max_id_width = max(max_id_width, len(str(entry.eid)))

        # Check total value width
        total_value_str = f"{total:.2f}"
        max_value_width = max(max_value_width, len(total_value_str))

    # Add some padding to ensure readability
    padding = 2
    return [
        max_name_width + padding,
        max_value_width + padding,
        max_date_percent_width + padding,
        max_id_width + padding,
    ]



"""Command line interface of financeager application."""

# PYTHON_ARGCOMPLETE_OK
import argparse
import os
import sys
import time
from datetime import datetime
from importlib.metadata import entry_points

import argcomplete
from dateutil import parser as du_parser
from rich.console import Console

import financeager

from . import (
    DEFAULT_TABLE,
    POCKET_DATE_FORMAT,
    RECURRENT_TABLE,
    UNSET_INDICATOR,
    __version__,
    clients,
    config,
    entries,
    exceptions,
    init_logger,
    listing,
    make_log_stream_handler_verbose,
    setup_log_file_handler,
)
from .pocket import FREQUENCY_CHOICES
from .server import pocket_names

logger = init_logger(__name__)

# Exit codes
SUCCESS = 0
FAILURE = 1


def main():
    """Main command line entry point of the application.

    The log directory is created. A FileHandler is added to the package logger.
    Available plugins are loaded.
    The program configuration is loaded.
    Relevant command line arguments and options are parsed and passed to
    'run()'.
    """
    os.makedirs(financeager.DATA_DIR, exist_ok=True)
    os.makedirs(financeager.CACHE_DIR, exist_ok=True)

    # Adding the FileHandler here avoids cluttering the log during tests
    setup_log_file_handler()

    group = "financeager.services"
    try:
        plugins = [ep.load()() for ep in entry_points().select(group=group)]
    except AttributeError:
        # Python 3.9
        plugins = [ep.load()() for ep in entry_points().get(group, [])]

    args = _parse_command(plugins=plugins)
    try:
        configuration = config.Configuration(
            args.pop("config_filepath"), plugins=plugins
        )
        exit_code = run(configuration=configuration, plugins=plugins, **args)

    except exceptions.InvalidConfigError as e:
        logger.critical(f"Invalid configuration: {e}")
        exit_code = FAILURE

    sys.exit(exit_code)


def run(command, configuration, plugins=None, verbose=False, sinks=None, **params):
    """Run 'command' request using additional 'params'.

    All 'params' except for formatting-related options are passed to
    'Client.safely_run()'.

    'configuration' is a config.Configuration object.

    'plugins' is an optional list of plugin.PluginBase objects that is forwarded
    to respective methods (e.g. ServicePlugins to clients.create()).

    If 'verbose' is set, debug level log messages are printed to the terminal.

    'sinks' is an optional Client.Sinks object to direct program output to.
    By default, the Python built-in logging interface is used.

    :return: UNIX return code (zero for success, non-zero otherwise)
    """
    if verbose:
        make_log_stream_handler_verbose()

    formatting_options = {}

    def _info(message):
        """Wrapper to format message and propagate it to stdout. The original
        message is logged at INFO-level.
        """
        response = _format_response(message, command, **formatting_options)
        logger.info(message)

        if isinstance(response, str):
            print(response)
        else:  # pragma: no cover
            Console().print(response)

    sinks = sinks or clients.Client.Sinks(_info, logger.error)

    try:
        _preprocess(params)
    except exceptions.PreprocessingError as e:
        sinks.error(e)
        return FAILURE

    formatting_options["default_category"] = configuration.get_option(
        "FRONTEND", "default_category"
    )
    if command == "list":
        # Extract formatting options; irrelevant, even confusing for Server
        for option in [
            "stacked_layout",
            "entry_sort",
            "category_sort",
            "category_percentage",
            "json",
        ]:
            formatting_options[option] = params.pop(option)
        if params["recurrent_only"]:
            formatting_options["recurrent_only"] = True

    exit_code = FAILURE
    client = clients.create(configuration=configuration, sinks=sinks, plugins=plugins)
    if client.safely_run(command, **params):
        exit_code = SUCCESS

    client.shutdown()

    return exit_code


def _preprocess(data):
    """Preprocess data to be passed to Client (e.g. convert date format, parse
    'filters' and 'month' options passed with list command).

    :raises: PreprocessError if preprocessing failed.
    """
    for field in ["date", "start", "end"]:
        date = data.get(field)
        if field == "end" and date == UNSET_INDICATOR:
            continue  # skip validation

        if date is not None:
            try:
                date = time.strftime(
                    POCKET_DATE_FORMAT,
                    du_parser.parse(date, yearfirst=True).timetuple(),
                )
                data[field] = date
            except ValueError:
                raise exceptions.PreprocessingError("Invalid date format.")

    filter_items = data.get("filters")
    parsed_items = {}
    if filter_items is not None:
        # convert list of "key=value" strings into dictionary
        try:
            for item in filter_items:
                key, value = item.split("=")
                parsed_items[key] = value.lower()

            for field, indicator in zip(
                ["category", "end"], [entries.CategoryEntry.DEFAULT_NAME, ""]
            ):
                # Substitute category default name, or empty string for end
                try:
                    if parsed_items[field] == indicator:
                        parsed_items[field] = None
                except KeyError:
                    # No field present
                    pass

            data["filters"] = parsed_items
        except ValueError:
            # splitting returned less than two parts due to missing separator
            raise exceptions.PreprocessingError(f"Invalid filter format: {item}")

    month = data.pop("month", None)
    if month is not None:
        today = datetime.today()
        if month == "current":
            date = today

        else:
            # Attempt to convert value for 'month' option (number (zero-padded
            # or not), or month name (abbreviated or full)) to 'filters'-option.
            # This is subject to the machine's locale setting
            date = None

            # -m / #m for parsing non-zero-padded numbers on UNIX / Windows
            for fmt in ["b", "B", "m", "-m", "#m"]:
                try:
                    date = datetime.strptime(month, f"%{fmt}").replace(year=today.year)
                    break
                except ValueError:
                    continue

            if date is None:
                raise exceptions.PreprocessingError(f"Invalid month: {month}")

        if filter_items is None:
            data["filters"] = {}

        # Overwrite 'filters' setting. Filter for entries of current year
        data["filters"]["date"] = f"{date.strftime('%Y-%m')}-"

    if any([data.get(f) for f in ["frequency", "start", "end"]]):
        # Assume that entry should be added to recurrent table
        data["table_name"] = RECURRENT_TABLE
    if any(n in parsed_items for n in ["start", "end", "frequency"]):
        # Assume that only recurrent table shall be displayed
        data["recurrent_only"] = True

    for field_name in ["category", "name"]:
        field = data.get(field_name)
        if field is not None:
            field = field.strip()
            if not len(field):
                raise exceptions.PreprocessingError(f"Empty {field_name} given.")
            # Update with sanitized data field
            data[field_name] = field


def _format_response(response, command, **listing_options):
    """Format the given response (dict or str) into human-readable text.
    If the response is a string, it is immediately returned.
    If the response does not contain any of the fields 'id', 'elements',
    'element', or 'pockets', an empty string is returned.
    The 'listing_options' are passed to listing.prettify().

    :return: str
    """
    if isinstance(response, str):
        return response

    eid = response.get("id")
    if eid is not None:
        verb = {
            "add": "Added",
            "update": "Updated",
            "remove": "Removed",
            "copy": "Copied",
        }[command]
        return f"{verb} element {eid}."

    elements = response.get("elements")
    if elements is not None:
        return listing.prettify(elements, **listing_options)

    element = response.get("element")
    if element is not None:
        return entries.prettify(
            element, default_category=listing_options["default_category"]
        )

    pockets = response.get("pockets", [])
    return "\n".join([p for p in pockets])


def _parse_command(args=None, plugins=None):
    """Parse the given list of args and return the result as dict."""

    parser = argparse.ArgumentParser(
        description="An application "
        "that helps you administering your daily expenses and earnings."
    )

    parser.add_argument(
        "-V",
        "--version",
        action="version",
        version=f"financeager version {__version__}",
        help="display version info and exit",
    )  # pragma: no cover

    subparsers = parser.add_subparsers(
        dest="command",
        metavar="command",
    )
    subparsers.required = True

    categories = _read_categories_for_cli_completion()

    add_parser = subparsers.add_parser("add", help="add an entry to the database")

    add_parser.add_argument("name", help="entry name")
    add_parser.add_argument("value", type=float, help="entry value")
    add_parser.add_argument(
        "-c", "--category", default=None, help="entry category"
    ).completer = argcomplete.ChoicesCompleter(categories)
    add_parser.add_argument("-d", "--date", default=None, help="entry date")

    add_parser.add_argument(
        "-t",
        "--table-name",
        default=None,
        choices=[DEFAULT_TABLE, RECURRENT_TABLE],
        help="""table to add the entry to. With 'recurrent', specify at
least a frequency, start date and end date are optional. Default:
'standard'""",
    )
    add_parser.add_argument(
        "-f",
        "--frequency",
        choices=FREQUENCY_CHOICES,
        help="""frequency of recurrent entry. If specified, '--table-name=recurrent'
is assumed""",
    )
    add_parser.add_argument(
        "-s", "--start", default=None, help="start date of recurrent entry"
    )
    add_parser.add_argument(
        "-e", "--end", default=None, help="end date of recurrent entry"
    )

    get_parser = subparsers.add_parser(
        "get", help="show information about single entry"
    )
    get_parser.add_argument("eid", help="entry ID")
    get_parser.add_argument(
        "-t",
        "--table-name",
        default=None,
        choices=[DEFAULT_TABLE, RECURRENT_TABLE],
        help="Table to get the entry from. Default: 'standard'.",
    )

    remove_parser = subparsers.add_parser(
        "remove", help="remove an entry from the database"
    )
    remove_parser.add_argument("eid", help="entry ID")
    remove_parser.add_argument(
        "-t",
        "--table-name",
        default=None,
        choices=[DEFAULT_TABLE, RECURRENT_TABLE],
        help="Table to remove the entry from. Default: 'standard'.",
    )

    update_parser = subparsers.add_parser(
        "update", help="update one or more fields of an entry"
    )
    update_parser.add_argument("eid", type=int, help="entry ID")
    update_parser.add_argument(
        "-t",
        "--table-name",
        help="Table containing the entry. Default: 'standard'",
        choices=[DEFAULT_TABLE, RECURRENT_TABLE],
    )
    update_parser.add_argument("-n", "--name", help="new name")
    update_parser.add_argument("-v", "--value", type=float, help="new value")
    update_parser.add_argument("-c", "--category", help="new category").completer = (
        argcomplete.ChoicesCompleter(categories)
    )
    update_parser.add_argument(
        "-d", "--date", help="new date (for standard entries only)"
    )
    update_parser.add_argument(
        "-f",
        "--frequency",
        choices=FREQUENCY_CHOICES,
        help="new frequency (for recurrent entries only)",
    )
    update_parser.add_argument(
        "-s", "--start", help="new start date (for recurrent entries only)"
    )
    update_parser.add_argument(
        "-e", "--end", help="new end date (for recurrent entries only)"
    )

    copy_parser = subparsers.add_parser(
        "copy", help="copy an entry from one pocket to another, or within one pocket"
    )
    copy_parser.add_argument("eid", help="entry ID")
    copy_parser.add_argument(
        "-s",
        "--source",
        default=None,
        dest="source_pocket",
        help="pocket to copy the entry from (default: main pocket)",
    )
    copy_parser.add_argument(
        "-d",
        "--destination",
        default=None,
        dest="destination_pocket",
        help="pocket to copy the entry to (default: source pocket)",
    )
    copy_parser.add_argument(
        "-t",
        "--table-name",
        default=None,
        choices=[DEFAULT_TABLE, RECURRENT_TABLE],
        help="Table to copy the entry from/to. Default: 'standard'.",
    )

    list_parser = subparsers.add_parser(
        "list", help="list all entries in the pocket database"
    )
    list_parser.add_argument(
        "-f",
        "--filter",
        default=None,
        action="append",
        dest="filters",
        metavar="FILTER",
        help="filter for name, "
        "date and/or category substring, e.g. name=beer category=groceries. "
        "Can be specified multiple times (then filters add up)",
    )
    list_parser.add_argument(
        "-s",
        "--stacked-layout",
        action="store_true",
        help="if true, display earnings and expenses in stacked layout, "
        "otherwise side-by-side. Not effective with --recurrent-only or --json option",
    )
    list_parser.add_argument(
        "--entry-sort",
        choices=[
            "name",
            "value",
            "date",
            "eid",
            "category",
            "start",
            "end",
            "frequency",
        ],
        help="key to sort entries by. The latter four keys can only be applied "
        "in combination with the --recurrent-only option",
    )
    list_parser.add_argument(
        "--category-sort",
        choices=["name", "value"],
        help="key to sort categories by. Not effective in combination with "
        "--recurrent-only or --json option",
    )
    list_parser.add_argument(
        "-P",
        "--category-percentage",
        action="store_true",
        help="show only category entries incl. percentage. Not effective in "
        "combination with --recurrent-only or --json option",
    )
    list_parser.add_argument(
        "-m",
        "--month",
        nargs="?",
        const="current",
        help="show only entries of given month (default: current) in current year",
    )
    list_parser.add_argument(
        "-r",
        "--recurrent-only",
        action="store_true",
        help="show only recurrent entries",
    )
    list_parser.add_argument(
        "-j",
        "--json",
        action="store_true",
        help="return result in JSON format instead formatting into table. "
        "Helpful for processing data with jq or similar tools.",
    )

    subparsers.add_parser("pockets", help="list all pocket databases")

    # Extend with plugin parsers
    plugins = plugins or []
    for plugin in plugins:
        plugin.cli_options.extend(subparsers)

    # Add common options to subparsers
    for subparser in subparsers.choices.values():
        subparser.add_argument(
            "-C",
            "--config-filepath",
            default=(
                financeager.CONFIG_FILEPATH
                if os.path.exists(financeager.CONFIG_FILEPATH)
                else None
            ),
            help=f"path to config file. Default: {financeager.CONFIG_FILEPATH}",
        )
        subparser.add_argument(
            "--verbose", action="store_true", help="Be verbose about internal workings"
        )

        if subparser in [
            add_parser,
            get_parser,
            remove_parser,
            update_parser,
            list_parser,
        ]:
            subparser.add_argument(
                "-p", "--pocket", help="name of pocket to modify or query"
            ).completer = argcomplete.ChoicesCompleter(
                pocket_names(financeager.DATA_DIR)
            )

    for subparser in [
        add_parser,
        get_parser,
        remove_parser,
        update_parser,
        copy_parser,
    ]:
        subparser.add_argument(
            "-r",
            "--recurrent",
            action="store_true",
            help="""alias for '-t recurrent'. If the -t option is simultaneously
specified, the -r option is ignored""",
        )

    argcomplete.autocomplete(parser)
    parsed_args = vars(parser.parse_args(args=args))

    # Set table name if not specified
    recurrent = parsed_args.pop("recurrent", None)
    if recurrent and parsed_args["table_name"] is None:
        parsed_args["table_name"] = RECURRENT_TABLE

    return parsed_args


def _read_categories_for_cli_completion():
    try:
        fp = os.path.join(financeager.CACHE_DIR, financeager.CATEGORIES_CACHE_FILENAME)
        if not os.path.exists(fp):
            return []
        with open(fp) as f:
            return f.read().splitlines()
    except Exception as e:
        logger.debug(str(e))
        logger.warning("Error when trying to read category cache.")
        return []



import unittest

from financeager.entries import BaseEntry, CategoryEntry
from financeager.entries import prettify as prettify_entry


class BaseEntryTestCase(unittest.TestCase):
    def test_attributes(self):
        entry = BaseEntry(
            **{"name": "groceries", "value": 123.45, "date": "2000-08-10"}
        )

        self.assertEqual(entry.name, "groceries")
        self.assertAlmostEqual(entry.value, 123.45, places=5)
        self.assertEqual(entry.date, "00-08-10")
        self.assertEqual(entry.eid, 0)

    def test_leap_year_date(self):
        entry = BaseEntry(
            **{
                "name": "leap",
                "value": 1,
                "date": "2000-02-29",
            }
        )
        self.assertEqual(entry.date, "00-02-29")


class NegativeBaseEntryTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.entry = BaseEntry(
            **{"name": "vw bully", "value": -6000, "date": "2000-01-01"}
        )

    def test_name(self):
        self.assertEqual(self.entry.name, "vw bully")

    def test_value(self):
        self.assertEqual(self.entry.value, 6000)


class CategoryEntryTestCase(unittest.TestCase):
    def setUp(self):
        self.entry = CategoryEntry(name="gifts")

    def test_name(self):
        self.assertEqual(self.entry.name, "gifts")

    def test_value(self):
        self.assertEqual(self.entry.value, 0.0)


class LongNegativeCategoryEntryTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.entry = CategoryEntry(
            name="This is quite a LOOONG Category",
            entries=[BaseEntry("entry", -100, "2000-08-13")],
        )

    def test_name(self):
        self.assertEqual(self.entry.name, "this is quite a looong category")

    def test_value(self):
        self.assertEqual(self.entry.value, 100)


class PrettifyBaseEntryTestCase(unittest.TestCase):
    def setUp(self):
        self.element = dict(
            name="soccer shoes", value=-123.45, date="04-01", category="sport equipment"
        )

    def test_prettify(self):
        self.assertEqual(
            prettify_entry(self.element, default_category=CategoryEntry.DEFAULT_NAME),
            """\
Name    : Soccer Shoes
Value   : -123.45
Date    : 04-01
Category: Sport Equipment""",
        )

    def test_prettify_default_category(self):
        element = self.element.copy()
        element["category"] = None
        self.assertEqual(
            prettify_entry(element, default_category=CategoryEntry.DEFAULT_NAME),
            f"""Name    : Soccer Shoes
Value   : -123.45
Date    : 04-01
Category: {CategoryEntry.DEFAULT_NAME.capitalize()}""",
        )


class PrettifyRecurrentEntryTestCase(unittest.TestCase):
    def setUp(self):
        self.element = dict(
            name="retirement money",
            value=567.0,
            category="income",
            frequency="monthly",
            start="01-01",
            end="12-31",
        )

    def test_prettify(self):
        self.assertEqual(
            prettify_entry(self.element, default_category=CategoryEntry.DEFAULT_NAME),
            """\
Name     : Retirement Money
Value    : 567.0
Frequency: Monthly
Start    : 01-01
End      : 12-31
Category : Income""",
        )


if __name__ == "__main__":
    unittest.main()
